(in-package :nc)

(export '(connect disconnect disconnect-all talk listens reset-listen 
          agent-listening-queue empty-queue empty-queue-p queue-duration) :nc)

(defclass port ()
  ((map-agent-queues :initform (make-hash-table :test #'eq)))
)

(defclass iac-central ()
  ((map-agents-ports-in :initform (make-hash-table :test #'eq))
   (map-agents-ports-out :initform (make-hash-table :test #'eq))
   (ports :accessor ports 
          :initform 
          (make-array 16 :initial-contents (loop repeat 16 collect (make-instance 'port))))
   )
  )

(setf *iac-central* (make-instance 'iac-central))



; queues

(defclass  fifo ()
   ((head :initform (list nil) :accessor head :initarg :head :type list)
    (tail  :initform nil :accessor tail :initarg :tail :type list) 
    (fifo-length :initform 0 :accessor fifo-length :initarg :fifo-length :type integer) 
    (duration :initform 0 :accessor duration :initarg :duration :type integer)
    (lock  :initform (make-lock 'fifo-lock)  :accessor lock :initarg :lock :type ccl::lock)))

(defmethod initialize-instance ((self fifo) &rest initargs)
  (declare (ignore initargs))
  (call-next-method)
  (setf (slot-value self 'tail) (slot-value self 'head))
  self)

(defmethod enqueue ((self fifo) (object t))
   (with-lock-grabbed ((lock self))
     (setf (rest (tail self)) (cons object nil)
           (tail self) (rest (tail self)))
     (incf (fifo-length self))
     (incf (duration self) (om::duration object))))


(defmethod dequeue ((self fifo) )
   (with-lock-grabbed ((lock self))
     (if (null (rest (head self)))
       nil
       (let ((object  (first (rest (head self)))))
         (prog1 object
           (setf (rest (head self)) (cddr (head self)))
           (unless (rest (head self)) (setf (tail self) (head self)))
           (decf (fifo-length self) )
           (decf (duration self) (om::duration object)))))))


(defmethod empty-queue-p ((self fifo))
  (null (rest (head self))))

(defmethod empty-queue-p ((self T))
  T)



(defmethod empty-queue ((self fifo))
   (with-lock-grabbed ((lock self))
     (setf (head self) (list nil) (tail self) (head self) (fifo-length self) 0 (duration self) 0)))

(defmethod empty-queue ((self T)))

(defmethod queue-size ((self fifo))
   (with-lock-grabbed ((lock self))
     (fifo-length self)))

(defmethod om::duration ((self t))
  0)


(defmethod queue-duration ((self fifo)) (duration self))
(defmethod queue-duration ((self T)) 0)



; iac central and port macros

(defmacro make-lockable-queue () `(make-instance 'fifo))

(defmacro theport (n)
  `(aref (ports *iac-central*) ,n))

(defmacro in-ports-for (agent)
  `(gethash ,agent (slot-value *iac-central* 'map-agents-ports-in )))

(defmacro out-port-for (agent)
  `(gethash ,agent (slot-value *iac-central* 'map-agents-ports-out )))

(defmacro iac-connect  (agent tofrom port) 
  `(case  ,tofrom
     (:toport   (pushnew  ,port (gethash ,agent (slot-value *iac-central* 'map-agents-ports-in ))))
     (:fromport (setf (gethash ,agent (slot-value *iac-central* 'map-agents-ports-out )) ,port ))))

(defmethod iac-disconnect  (agent tofrom port)
  (let* ((htable (slot-value *iac-central* 
                             (case tofrom 
                               (:toport 'map-agents-ports-in)
                               (:fromport 'map-agents-ports-out )))))
    (case tofrom
      (:toport (let ((remaining (remove port (gethash agent htable))))
                 (if remaining 
                   (setf (gethash agent htable) remaining)
                   (remhash agent htable))))
      (:fromport (remhash agent htable)))))

(defmacro iac-disconnect-agent (agent)
  `(progn 
     (remhash ,agent (slot-value *iac-central* 'map-agents-ports-in ))
     (remhash ,agent (slot-value *iac-central* 'map-agents-ports-out ))))
     
(defmacro map-port-queues (port aqueue &body body)
  `(maphash #'(lambda (key ,aqueue) ,. body) (slot-value  port 'map-agent-queues )))

(defmacro add-port-queue (port agent)
  `(setf (gethash ,agent (slot-value (theport ,port) 'map-agent-queues ))
         (make-lockable-queue)))

(defmacro rem-port-queue (port agent)
  `(remhash ,agent (slot-value (theport ,port) 'map-agent-queues )))

(defmacro agent-queue (agent port)
  `(gethash ,agent (slot-value ,port 'map-agent-queues )))

;  connect agents

(defmethod connect ((agent agent) direction port)
  (iac-connect agent direction port)
  (when (eq direction :fromport) 
    (add-port-queue port agent)
    (setf (inport agent) port)))

(defmethod disconnect ((agent agent) direction port)
  (iac-disconnect agent direction port)
  (when (eq direction :fromport) 
    (rem-port-queue port agent)
    (setf (inport agent) nil)))

(defmethod disconnect-all ((agent agent))
  (when (out-port-for agent) (rem-port-queue (out-port-for agent) agent))
  (iac-disconnect-agent agent))


(defmethod agent-listening-queue ((self agent))
  (let ((port (out-port-for self)))
    (and port (agent-queue self (theport port)))))


; read or write on iac central

(defmethod talk ((agent agent) (data T))
  (loop for port in (in-ports-for agent)
        do (talk (theport port) data)))

(defmethod listens ((agent agent))
  (let ((port (out-port-for agent)))
    (and port
         ; right now we assume that an agent may listen only one port at a time
       (listen-for agent  (theport port)))))

(defmethod reset-listen ((self agent))
  (let ((port (out-port-for self)))
    (and port
       (empty-queue (agent-queue self (theport port))))))



; read or write on a port

(defmethod talk ((port port) (data T))
  (map-port-queues  port queue
    (enqueue queue data)))

(defmethod listen-for ((agent agent) (port port))
  (let ((queue (agent-queue agent port)))
    (unless (empty-queue-p queue) (dequeue queue))))



#|


(progn
(defclass talker (agent)
  ((counter :accessor counter :initform 0)))

(defclass talker2 (agent)
  ((counter :accessor counter :initform 0)))


(defclass receive (agent)
  ((check :accessor check :initform 0)))

(defmethod task ((self talker))
  ;(call-next-method)
  (talk self (incf (counter self)))
  (sleep 0.5)
  )

(defmethod task ((self talker2))
  ;(call-next-method)
  (talk self (decf (counter self)))
  (sleep 0.5)
  )

(defmethod task ((self receive))
  ;(call-next-method)
  (setf (check  self) (listens self))
  (sleep 0.8))


(setf tt (NewAgent 'talker)) (connect tt :toport 0) 
(setf rr (NewAgent 'receive))(connect rr :fromport 0)
(setf tt1 (NewAgent 'talker2))
(setf rr1 (NewAgent 'receive))
(setf tt2 (NewAgent 'talker))
(connect tt :toport 5)
(connect tt1 :toport 0)
(connect rr1 :fromport 5)
(connect tt2 :toport 5)

)


(kill tt)(setf tt (NewAgent 'talker)) (connect tt :toport 0) (inspect (agent-queue rr (theport 0)))

(kill tt)
(kill rr)

(mapc 'kill (list tt rr tt1 rr1 tt2))

(reset-listen rr1)

(in-ports-for tt2)
(in-ports-for rr1)
(out-port-for rr1)
(out-port-for tt)
(disconnect-all tt)
(disconnect-all rr)

(inspect (agent-queue rr1 (theport 5)))
(inspect *iac-central*)


)



(

(inspect rr)

(kill tt)



  
|#