;;-----------------------------------------------------------------------------
;; Moscou
;;
;; Basic OpenSoundControl (OSC) for OpenMusic. Based on early work done by
;; Carlos Agon.
;;
;; Author : Carl Seleborg
;; Date : 2004
;; Version : 1.0
;; Copyright Ircam 2004
;;-----------------------------------------------------------------------------

(in-package :om)



;;-----------------------------------------------------------------------------
;; The Moscou:receive method.
;;
;; This method works a little differently from other methods. When the user
;; evaluates it, it's border turns blue (or some other system-dependent colour)
;; indicating that the method is "listening" for OSC messages on the given
;; port.
;;-----------------------------------------------------------------------------
#|
(defmethod osc-receive (port patch)
  t)



;;-----------------------------------------------------------------------------
;; We need to define a class that will contain the properties of an instance of
;; Moscou:receive box, since each individual box is associated with a background
;; process that sits and waits for OSC messages to arrive.
;;-----------------------------------------------------------------------------
(defclass ReceiveBox (OMBoxCall) 
   ((etat     :initform nil  :initarg :etat      :accessor etat)
    (process  :initform nil  :initarg :process   :accessor process)
    (endpoint :initform nil  :initarg :endpoint  :accessor endpoint)
    (buf      :initform nil  :initarg :buf       :accessor buf)))



;;-----------------------------------------------------------------------------
;; We then have to redefine a few functions that implement the behaviour of the
;; box. Here, we replace the standard frame class with our freshly defined
;; box class.
;;-----------------------------------------------------------------------------
(defmethod get-frame-class ((self ReceiveBox)) 'ReceiveBoxFrame)



;;-----------------------------------------------------------------------------
;; Call the Moscou:receive method when the box is evaluated.
;;-----------------------------------------------------------------------------
(defmethod OM::get-boxcallclass-fun ((self (eql 'osc-receive))) 'ReceiveBox)



;;-----------------------------------------------------------------------------
;; There is no editor for our box.
;;-----------------------------------------------------------------------------
(defmethod OpenEditorframe ((self ReceiveBox)) 
   (not (dialog-message "Compiled Function receive")))



;;-----------------------------------------------------------------------------
;; Our box cannot be locked, since it has no result. What is important
;; is what happens during it's execution, not the result it returns.
;;-----------------------------------------------------------------------------
(defmethod allow-lock-button ((self ReceiveBox))  nil)



;;-----------------------------------------------------------------------------
;; (Not really sure what this does. Check with Carlos for that.)
;;-----------------------------------------------------------------------------
(defmethod call-gen-code ((self ReceiveBox) numout)
   (declare (ignore numout))
   `(if ,(gen-code (first (inputs self)) 0) ,(gen-code (second (inputs self)) 0) ,(gen-code (third (inputs self)) 0)))



;;-----------------------------------------------------------------------------
;; (Not really sure what this does. Check with Carlos for that.)
;;-----------------------------------------------------------------------------
(defmethod gen-code-call ((self ReceiveBox))
   (call-gen-code self 0))
 


;;-----------------------------------------------------------------------------
;; This is the function actually executed when the user evaluates the box.
;; It basically starts a process waiting for OSC messages.
;;-----------------------------------------------------------------------------
(defmethod omNG-box-value ((self ReceiveBox) &optional (numout 0))
  (declare (ignore numout))
   (setf (etat self) (not (etat self)))
   (if (etat self)
     (let* ((args  (eval-box-inputs self))
            patch)
       (when (second args)
         (progn (setf patch (first (connected? (nth 1 (inputs self)))))
                (setf patch (reference patch))
                (unless (patch-p patch)
                  (error "Second input must be a function symbol or a patch"))))
       (setf (buf self) (om-make-pointer *max-buf-size* t))
       (setf (endpoint self) (om-make-receive-connection (first args)))
       (setf (process self) 
         (om-run-process (name self) 'start-OSC-receive (list (endpoint self) (buf self) nil patch)))
       )
     (kill-receive (process self) (endpoint self) (buf self)))
  (when  (car (frames self))
    (om-invalidate-view (car (frames self)) t))
   t)



;;-----------------------------------------------------------------------------
;; We need to draw a special frame around our receive box, so that the user
;; sees which state it is in (listening or idle).
;;-----------------------------------------------------------------------------
(omg-defclass ReceiveBoxFrame (BoxFrame) ())

;;-----------------------------------------------------------------------------
;; Redefinition of the draw-after-box method.
;;-----------------------------------------------------------------------------
(defmethod draw-after-box ((self ReceiveBoxFrame)) 
   (when (etat (object self))
     (om-with-focused-view self
       (om-draw-hilite-rect 0 0 (w self) (h self)))))




|#
;;-----------------------------------------------------------------------------
;;-----------------------------------------------------------------------------
;;-----------------------------------------------------------------------------




;;-----------------------------------------------------------------------------
;; Wait for an OSC bundle to arrive.
;;-----------------------------------------------------------------------------
(defun start-OSC-receive (endpoint buffer host patch)
  (let ((result nil))
    (loop while t do
          (handler-case
            (let ((bufsize (om-receive-udp-packet endpoint buffer *max-buf-size*))
                  bundle)
              (setf bundle (parse-buffer buffer 0 bufsize))
              ;(om-print (list "RECEIVED " bundle))
              (setf result (deliver-bundle bundle patch))
              )

            (error (any)
                   (error any)
                   )))))



;;-----------------------------------------------------------------------------
;; Kill the receiving process.
;;-----------------------------------------------------------------------------
(defun kill-receive (process endpoint my-buf )
  (om-kill-process process)
  (om-close-udp-connection endpoint)
  (om-free-pointer my-buf))



;;-----------------------------------------------------------------------------
;; Act when receiving a bundle.
;;
;; We provide a real bad-boy feature here : if a message is prefixed with the
;; selector "/lisp", the rest of the message is executed as a normal lisp
;; function call.
;;-----------------------------------------------------------------------------
(defun deliver-bundle (bundle patch)
  (loop for item in bundle collect
        (cond
         ((stringp (car item)) (cond ((string-equal (car item) "/lisp") (execute-message item))
                                     ((string-equal (car item) "/lisp-string") (execute-message-from-string item))
                                     (t (deliver-message item patch))))
         ((consp (car item)) (deliver-bundle item patch))
         (t nil))
  )
)


#|
(defun deliver-message (message patch)
  (apply (intern (string (code patch)) :om) (list message)))

|#


(defun execute-message (message)
  (apply (if (symbolp (cadr message)) (cadr message) (read-from-string (cadr message))) (cddr message)))

(defun execute-message-from-string (message)
  ; (print (list "execute-message-from-string: Message received : " message))
  (eval (read-from-string (cadr message)))
)


;;-----------------------------------------------------------------------------
;; Low-level parsing routines.
;;-----------------------------------------------------------------------------

(defun advance-str (str i)
  (+ i (length str) (- 4 (mod (length str) 4))))

(defun read-int32 (buff i)
  (let ((a (om-read-ptr buff i :unsigned-byte))
        (b (om-read-ptr buff (+ i 1) :unsigned-byte))
        (c (om-read-ptr buff (+ i 2) :unsigned-byte))
        (d (om-read-ptr buff (+ i 3) :unsigned-byte)))
    (logior (ash a 24) (ash b 16) (ash c 8) d)))

(defun read-f32 (buff i)
  (om-read-ptr buff i :single-float))

(defun read-osc-string (buff i)
  (om-read-ptr buff i :string))

(defun parse-buffer (buf i size)
  (let* ((head (read-osc-string buf i)) rep)
    (setf i (advance-str head i))
    (if (string-equal "#bundle" head)
      (let (time1 time2 elemsize mssg)
        (setf time1 (read-int32 buf i))
        (setf i (+ i 4))
        (setf time2 (read-int32 buf i))
        (setf i (+ i 4))
        (setf elemsize (read-int32 buf i))
        (setf i (+ i 4))
        (loop while (< i size) do
              (setf mssg (read-osc-string buf i))
              (setf i (advance-str mssg i))
              (if (string-equal "#bundle" mssg)
                (push  (parse-bndl buf i elemsize) rep)
                (push  (parse-mssg buf i mssg elemsize) rep)
                )
              (setf i (+ i elemsize))
              (setf elemsize (read-int32 buf i))
              (setf i (+ i 4)))
        (reverse rep))
      (om-beep-msg "sorry i receive only bundles"))))



(defun parse-bndl (buf i size)
  (let (time1 time2 elemsize mssg rep)
    (setf time1 (read-int32 buf i))
    (setf i (+ i 4))
    (setf time2 (read-int32 buf i))
    (setf i (+ i 4))
    (setf elemsize (read-int32 buf i))
    (setf i (+ i 4))
    (loop while (< i size) do
          (setf mssg (read-osc-string buf i))
          (setf i (advance-str mssg i))
          (if (string-equal "#bundle" mssg)
            (push  (parse-bndl buf i elemsize) rep)
            (push  (parse-mssg buf i mssg elemsize) rep))
          (setf i (+ i elemsize))
          (setf elemsize (read-int32 buf i))
          (setf i (+ i 4)))
    (reverse rep)))

(defun parse-mssg (buf i head size)
  (let (typestring args)
    (setf typestring (read-osc-string buf i))
    (setf i (advance-str typestring i))
    (loop for k from 1 to (- (length typestring) 1) do
          (let ((char (elt typestring k)))
            (cond
             ((equal char #\f) 
              (push (read-f32 buf i) args )
              (setf i (+ i 4)))
             ((equal char #\i) 
              (push (read-int32 buf i) args )
              (setf i (+ i 4)))
             ((equal char #\s) 
              (let ((string (read-osc-string buf i)))
                (push string args)
                (setf i (advance-str string i))))
             ((equal char #\b) (om-print "parse-mssg: blob type not handled yet... Sorry!"))
             (t (error (format nil "no conozco el tipo ~S" char))))))
    (cons head (reverse args) )))








