;;-----------------------------------------------------------------------------
;; Moscou
;;
;; Basic OpenSoundControl (OSC) for OpenMusic. Based on early work done by
;; Carlos Agon.
;;
;; Author : Carl Seleborg
;; Date : 2004
;; Version : 1.0
;; Copyright Ircam 2004
;;
;; Notes :
;; 
;; * Sending non-standard data types
;;   -------------------------------
;;   There is a problem sending non-standard data types using the currently
;;   available version of the OSC client library, due to very limited type
;;   verification. Currently, types are checked agains the type-tag string
;;   upon adding the parameter to the message, which means that only the few
;;   existing functions for adding parameters can build a message without
;;   errors. In the code, you will find an out-commented atempt to add lists to
;;   OSC messages, but this is not possible for the same reasons : the type
;;   checking mechanism is extremely limited in the OSC client library. 
;;-----------------------------------------------------------------------------

(in-package :om)

;;-----------------------------------------------------------------------------
;; The OpenMusic send function. This is for the visual programming style.
;; If you're an ascii-addict, check out send-osc-bundle.
;;-----------------------------------------------------------------------------
(om::defmethod osc-send (bundle host port)
  (send-osc-bundle port host bundle))


;;-----------------------------------------------------------------------------
;; Writes an argument value into an OSCBuffer structure.
;;-----------------------------------------------------------------------------
(defun write-osc-arg (arg OSCbufp)
  ; (print (list "write-osc-arg: Adding " arg " to buffer " OSCbufp))
  (cond 
   ((integerp arg)   (osc::OSC_writeIntArg OSCbufp arg))
   ((floatp arg)     (osc::OSC_writeFloatArg OSCbufp arg))
   ((symbolp arg)    (OSC-write-String-Arg OSCbufp (string-downcase (string arg))))
   ((stringp arg)    (OSC-write-String-Arg OSCbufp arg))
   ;((listp arg) (loop for item in arg do
   ;                   (write-osc-arg item OSCbufp)))
   (t (error "unknown argument type"))))


;;-----------------------------------------------------------------------------
;; Builds a type string according to OSC specifications. Only three basic types
;; are currently recognized : integers, floats and strings. Blobs are not yet
;; handled, nor is any of the additional type tags defined by the OSC
;; specification.
;;-----------------------------------------------------------------------------
(defun get-soc-types (args &optional (with-initial-comma t))
  (let ((rep (if with-initial-comma "," "")))
    (loop for arg in args do
          (cond 
           ((integerp arg) (setf rep (om::string+ rep "i")))
           ((floatp arg) (setf rep (om::string+ rep "f")))
           ((symbolp arg) (setf rep (om::string+ rep "s")))
           ((stringp arg) (setf rep (om::string+ rep "s")))
           ; ((listp arg) (setf rep (om::string+ rep "[" (get-soc-types arg nil) "]")))
           (t (error "unknown argument type"))))
    rep))


;;-----------------------------------------------------------------------------
;; Writes a bundle into an OSCBuffer structure. There is room for a time-tag
;; parameter, but it is currently ignored, and automatically set to
;; "immediatly" in the message, meaning that there is no official way to
;; schedule the effect of an OSC message in the future.
;;-----------------------------------------------------------------------------
(defun write-a-bundle (bundle OSCbufp tt)   ;; The tt parameter is unused : we use the "immediatly" time tag for now.
  (osc::OSC_openBundle OSCbufp 0 1)
  (loop for args in bundle do
        (cond
         ((consp (car args)) (write-a-bundle args oscbufp tt))
         ((or (stringp (car args)) (symbolp (car args)))
          (let ((types (get-soc-types (cdr args)))
                (name (if (stringp (car args)) (car args) (string-downcase (string (car args))))))
            (OSC-write-Address-and-Types OSCbufp name types)
            (loop for item in (cdr args) do 
                  (write-osc-arg item OSCbufp))))
        (t (error "bad syntax"))))
  (osc::OSC_closeBundle OSCbufp))


;;-----------------------------------------------------------------------------
;; Send an OSC bundle to a given host on a given port. The function currently
;; uses UDP for this, courtesy of Mr OpenTransport.
;;
;; This is the function you want to call if you want to send OSC messages
;; from your Lisp code. For instance, you can do :
;;
;;        (send-osc-bundle 3001 "127.0.0.1" '(("/foo" 330.0 3 "casa")))
;;
;; Note that we only send bundles, meaning that you must send a list of flat
;; lists. Other formats will not work, especially nested lists, for reasons
;; explained at the top of this file.
;;-----------------------------------------------------------------------------
(defun send-osc-bundle (port host bundle &optional (timetag 0.))
  (let* ((conn (om-make-send-connection host port))
         (oscbufp (om-make-pointer *max-buf-size*))
         (ptr (om-make-pointer *max-buf-size*)))
    (handler-case
        (let ((tt (osc-make-time-tag 0 1))) 
          (osc::OSC_initBuffer OSCbufp *max-buf-size* ptr)
          (write-a-bundle bundle oscbufp tt)
          (om-send-udp-packet conn (osc::OSC_getPacket OSCbufp) (osc::OSC_packetSize OSCbufp))
          (om-close-udp-connection conn)
          (osc::OSC_resetBuffer OSCbufp)
          (om-free-pointer ptr) t)
      
      (error (any)
        (om-close-udp-connection conn)
        (osc::OSC_resetBuffer OSCbufp)
        (om-free-pointer ptr)
        (error any)))))


;;-----------------------------------------------------------------------------
;; Debugging convenience that prints out the bytes of a memory buffer.
;;-----------------------------------------------------------------------------
(defun print-buffer (buf size)
  (print (loop for i from 0 to (- size 1) collect
               ;(CCL::%CODE-CHAR (cl-user::%get-byte buf i))
               (code-char (om-read-ptr buf i :unsigned-byte))
               )))



 