;;; OpenGL for MCL Camera 
;;; Alexander Repenning, 8/27/01
;;; updated: 12/6/02 ;; sustain animation


(in-package :ccl)

(defclass CAMERA ()
  ((view :accessor view :initarg :view :documentation "view or window")
   (eye-x :accessor eye-x :initform 0.0)
   (eye-y :accessor eye-y :initform 0.0)
   (eye-z :accessor eye-z :initform 2.0)
   (center-x :accessor center-x :initform 0.0)
   (center-y :accessor center-y :initform 0.0)
   (center-z :accessor center-z :initform 0.0)
   (up-x :accessor up-x :initform 0.0)
   (up-y :accessor up-y :initform 1.0)
   (up-z :accessor up-z :initform 0.0)
   (fovy :accessor fovy :initform 60.0 :documentation "viewing angle")
   (aspect :accessor aspect :initform 1.0)
   (near :accessor near :initform 0.005)
   (far :accessor far :initform 2000.0))
  (:documentation "OpenGL camera to control user perspective"))
   

;________________________________
; Specification                  |
;_______________________________/


(defgeneric AIM-CAMERA (Camera &key Eye-X Eye-Y Eye-Z Center-X Center-Y Center-Z 
                               Up-X Up-Y Up-Z
                               Fovy Aspect Near Far Animation-Time)
  (:documentation "Set the aim of the camera. Use Animation-Time to make a smoot transition"))


(defgeneric TRACK-MOUSE-3D (Camera)
  (:documentation "Track the mouse to control the aim of the camera"))

;________________________________
; Implementation                 |
;_______________________________/

(defmethod AIM-CAMERA ((Self camera) &key Eye-X Eye-Y Eye-Z
                         Center-X Center-Y Center-Z 
                         Up-X Up-Y Up-Z
                         Fovy Aspect Near Far
                         Animation-Time)
  (declare (ignore Animation-Time))
  ;; for now ignore Animation Time => no smoot transition, just instant warp
  (setf (eye-x Self) (or Eye-X (eye-x Self)))
  (setf (eye-y Self) (or Eye-Y (eye-y Self)))
  (setf (eye-z Self) (or Eye-Z (eye-z Self)))
  (setf (center-x Self) (or Center-X (center-x Self)))
  (setf (center-y Self) (or Center-Y (center-y Self)))
  (setf (center-z Self) (or Center-Z (center-z Self)))
  (setf (up-x Self) (or Up-x (up-x Self)))
  (setf (up-y Self) (or Up-y (up-y Self)))
  (setf (up-z Self) (or Up-z (up-z Self)))
  (setf (fovy Self) (or Fovy (fovy Self)))
  (setf (aspect Self) (or Aspect (aspect Self)))
  (setf (near Self) (or Near (near Self)))
  (setf (far Self) (or Far (far Self)))
  ;; set the AGL context
  (aglSetCurrentContext (aglcontext (view Self)))
  (agl-error-check)
  ;; set the Transformation matrix
  (glmatrixmode gl_projection)
  (glloadidentity)
  (gluperspective (fovy Self) (aspect Self) (near Self) (far Self))
  (glmatrixmode gl_modelview)
  (glloadidentity)
  (gluLookat 
   (eye-x Self) (eye-y Self) (eye-z Self)
   (center-x Self) (center-y Self) (center-z Self) 
   (up-x Self) (up-y Self) (up-z Self))
  (invalidate-view (view Self)))

;________________________________
; Rotation Matrix Functions      |
;_______________________________/

(defun ROTATE-X (x y z Phy)
  (let ((Sin (sin Phy)) (Cos (cos Phy)))
    (values
     x
     (- (* y Cos) (* z Sin))
     (+ (* y Sin) (* z Cos)))))


(defun ROTATE-Y (x y z Phy)
  (let ((Sin (sin Phy)) (Cos (cos Phy)))
    (values
     (+ (* x Cos) (* z Sin))
     y
     (- (* z Cos) (* x Sin)))))


(defun ROTATE-Z (x y z Phy)
  (let ((Sin (sin Phy)) (Cos (cos Phy)))
    (values
     (- (* x Cos) (* y Sin))
     (+ (* x Sin) (* y Cos))
     z)))
;________________________________
; Camera Mouse Control           |
;_______________________________/

(defmethod TRACK-MOUSE-3D ((Self camera))
  (cond 
   ((option-key-p) (track-mouse-zoom Self (if (shift-key-p) 0.005 0.02)))
   ((command-key-p) (track-mouse-pan Self (if (shift-key-p) 0.001 0.01)))
   (t (track-mouse-spin Self (if (shift-key-p) 0.01 0.04)))))


(defmethod TRACK-MOUSE-ZOOM ((Self camera) Gain)
  (declare (ignore x))
  (let* ((Eye-x (eye-x Self)) (Eye-y (eye-y Self)) (Eye-z (eye-z Self))
         (Center-x (center-x Self)) (Center-y (center-y Self)) (Center-z (center-z Self))
         (vx (- Center-x Eye-x))
         (vy (- Center-y Eye-y))
         (vz (- Center-z Eye-z))
         (Length (sqrt (+ (* vx vx) (* vy vy) (* vz vz))))
         (Distance 0)
         (y0 (point-v (view-mouse-position (view Self)))))
    (setq vx (/ vx Length))
    (setq vy (/ vy Length))
    (setq vz (/ vz Length))
    (loop
      (unless (mouse-down-p) (return))
      ;; sustain animation
      (animate (view Self) (delta-time (view Self)))
      (view-draw-contents (view Self))
      ;; adjust camera
      (let* ((Position (view-mouse-position (view Self)))
             (New-Distance (* (- (point-v Position) y0) Gain)))
        (unless (= New-Distance Distance)
          (aim-camera Self :eye-x (+ Eye-x (* vx Distance))
                      :eye-y (+ Eye-y (* vy Distance))
                      :eye-z (+ Eye-z (* vz Distance)))
          (view-draw-contents (view Self))
          (setq Distance New-Distance))))))


(defmethod TRACK-MOUSE-PAN ((Self camera) Gain)
  (let ((Eye-x (eye-x Self)) (Eye-y (eye-y Self)) 
        (Center-x (center-x Self)) (Center-y (center-y Self))
        (Old-dx 0) (Old-dy 0))
    (let* ((Position (view-mouse-position (view Self)))
           (x0 (point-h Position))
           (y0 (point-v Position)))
      (loop
        (unless (mouse-down-p) (return))
        ;; sustain animation
        (animate (view Self) (delta-time (view Self)))
        (view-draw-contents (view Self))
        ;; adjust camera
        (let* ((Position (view-mouse-position (view Self)))
               (dx (* (- (point-h Position) x0) Gain))
               (dy (* (- (point-v Position) y0) Gain)))
          (unless (and (= dx Old-dx) (= dy Old-dy))
            ;; Pan needs to be orthogonal to view vector to make sense
            (aim-camera Self :eye-x (- Eye-x dx) :eye-y (+ Eye-y dy)
                        :center-x (- Center-x dx) :center-y (+ Center-y dy))
            (view-draw-contents (view Self))
            (setq Old-dx dx)
            (setq Old-dy dy)))))))


(defmethod TRACK-MOUSE-SPIN ((Self camera) Gain)
  (let* ((Eye-x (eye-x Self)) (Eye-y (eye-y Self)) (Eye-z (eye-z Self))
         (Center-x (center-x Self)) (Center-y (center-y Self)) (Center-z (center-z Self))
         (vx (- Eye-x Center-x))
         (vy (- Eye-y Center-y))
         (vz (- Eye-z Center-z))
         (Old-Head 0)
         (Old-Pitch 0)
         (Position (view-mouse-position (view Self)))
         (x0 (point-h Position))
         (y0 (point-v Position)))
    (loop
      (unless (mouse-down-p) (return))
      ;; sustain animation
      (animate (view Self) (delta-time (view Self)))
      (view-draw-contents (view Self))
      ;; adjust camera
      (let* ((Position (view-mouse-position (view Self)))
             (Head (* (- x0 (point-h Position)) Gain))
             (Pitch (* (- y0 (point-v Position)) Gain)))
        (unless (and (= Head Old-Head) (= Pitch Old-Pitch))
          (multiple-value-bind (x1 y1 z1)
                               (rotate-y vx vy vz Head)
            (multiple-value-bind (x2 y2 z2)
                                 (rotate-x x1 y1 z1 Pitch)
              ;; need to adjust also the head if the pitch has changed
              ;; this code does not work yet properly
              (multiple-value-bind (xh yh zh)
                                   (rotate-y 0.0 1.0 0.0 Head)
                (multiple-value-bind (up-x up-y up-z)
                                     (rotate-x xh yh zh Pitch)

              (aim-camera Self :eye-x (+ Center-x x2)
                          :eye-y (+ Center-y y2)
                          :eye-z (+ Center-z z2)
                          :up-x up-x
                          :up-y up-y
                          :up-z up-z)
              (view-draw-contents (view Self))
              (setq Old-Head Head)
              (setq Old-Pitch Pitch))))))))))
      
      

#| Examples:

(defparameter *Window* (make-instance 'opengl-window))


|#

