;;; OpenGL for MCL Game Engine
;;; Alexander Repenning 11/20/02
;;; Updated: 12/06/02

(in-package :ccl)


(defclass GAME-WINDOW (opengl-window)
  ((agents :accessor agents :initform nil :initarg :agents)))


(defmethod INIT ((Self game-window))
  ;; GL setup
  (glClearColor 0.5 0.5 0.5 0.0)
  (glShadeModel gl_smooth)
  (glPixelStorei gl_unpack_alignment 1)
  ;; define material
  (with-rgba-vector Specular (0.5 0.5 0.5 0.0)
    (glmaterialfv gl_front_and_back gl_specular Specular))
  (glmaterialf gl_front_and_back gl_shininess 120.0)
  (with-rgba-vector V (1.0 1.0 1.0 1.0)
    (glmaterialfv gl_front_and_back gl_ambient_and_diffuse V))
  ;; light
  (with-rgba-vector Position (3.0 3.0 3.0 1.0)
    (gllightfv gl_light0 gl_position Position))
  (with-rgba-vector White-Light (1.0 1.0 1.0 1.0)
    (gllightfv gl_light0 gl_diffuse White-Light))
  (with-rgba-vector Specular-Light (1.0 1.0 1.0 1.0)
    (gllightfv gl_light0 gl_specular Specular-Light))
  ;; use wood like material
  (with-rgba-vector V (#.(/ 231 255.0) #.(/ 156 255.0) #.(/ 109 255.0) 1.0)
    (glmaterialfv gl_front_and_back gl_ambient_and_diffuse V))
  ;; enablers
  (glenable gl_lighting)
  (glenable gl_light0)
  (glenable gl_depth_test)
  (gllinewidth 5.0)
  (gledgeflag gl_false)
  (glenable gl_line_smooth)             ;; anti aliased lines
  (glEnable gl_point_smooth)
  ;; Textures
  (glenable gl_texture_2d)
  (gltexenvf gl_texture_env gl_texture_env_mode gl_decal)
  (glpointsize 9.0)
  ;; Alpha
  (glenable gl_blend)
  (glBlendFunc gl_src_alpha gl_one_minus_src_alpha)
  ;; camera
  (aim-camera (camera Self) :eye-z 40.0 :near 1.0)
  ;; compose the scene
  (compose-scene Self))


(defmethod COMPOSE-SCENE ((Self game-window))
  (setf (agents Self) nil)
  ;; Sky dome for orientation
  (attach Self (make-instance 'sky-dome :name "The World around me" :pitch -90.0 :view Self)))


(defmethod VIEW-KEY-EVENT-HANDLER ((Self game-window) Key)
  ;; pass key events to agents
  (dolist (Agent (agents Self)) (key-event-handler Agent Key)))


(defmethod ATTACH ((Self game-window) Agent)
  ;; attach agents to this window
  (push Agent (agents Self)))


(defmethod DISPLAY ((Self game-window))
  (glClear (logior gl_color_buffer_bit gl_depth_buffer_bit))
  ;; draw all the bones
  (dolist (Agent (agents Self)) (display Agent)))


(defmethod ANIMATE ((Self game-window) Time)
  ;; animate all agents
  (dolist (Agent (agents Self)) (animate Agent Time)))

;_______________________________________
;  Agents                               |
;_______________________________________

(defclass AGENT-3D ()
  ((view :accessor view :initform nil :initarg :view)
   (name :accessor name :initform "untitled" :initarg :name)
   (x :accessor x :initform 0.0s0 :initarg :x)
   (y :accessor y :initform 0.0s0 :initarg :y)
   (z :accessor z :initform 0.0s0 :initarg :z)
   (scale-x :accessor scale-x :initform 1.0s0 :initarg :scale-x)
   (scale-y :accessor scale-y :initform 1.0s0 :initarg :scale-y)
   (scale-z :accessor scale-z :initform 1.0s0 :initarg :scale-z)
   (x-turn :accessor x-turn :initform 0.0s0 :initarg :x-turn)
   (y-turn :accessor y-turn :initform 0.0s0 :initarg :y-turn)
   (z-turn :accessor z-turn :initform 0.0s0 :initarg :z-turn)
   (velocity-x :accessor velocity-x :initform 0.0s0 :initarg :velocity-x)
   (velocity-y :accessor velocity-y :initform 0.0s0 :initarg :velocity-y)
   (velocity-z :accessor velocity-z :initform 0.0s0 :initarg :velocity-z)
   (acceleration-x :accessor acceleration-x :initform 0.0s0 :initarg :acceleration-x)
   (acceleration-y :accessor acceleration-y :initform 0.0s0 :initarg :acceleration-y)
   (acceleration-z :accessor acceleration-z :initform 0.0s0 :initarg :acceleration-z)
   (roll :accessor roll :initform 0.0s0 :initarg :roll)
   (pitch :accessor pitch :initform 0.0s0 :initarg :pitch)
   (heading :accessor heading :initform 0.0s0 :initarg :heading)
   (agents :accessor agents :initform nil :initarg :agents)))


(defmethod DISPLAY :before ((Self agent-3d))
  ;; BEGIN display with transformations
  (glPushmatrix)
    ;; Scale
    (glScalef (scale-x Self) (scale-y Self) (scale-z Self))
    ;; translate to origin
    ;;(glTranslatef (- (x Self)) (- (y Self)) (- (z Self)))
    (glTranslatef (x Self) (y Self) (z Self))
    (glRotatef (roll Self) 0.0s0 0.0s0 1.0s0)
    (glRotatef (heading Self) 0.0s0 1.0s0 0.0s0)
    (glRotatef (pitch Self) 1.0s0 0.0s0 0.0s0))


(defmethod DISPLAY ((Self agent-3d))
  ;; display all subagents
  (dolist (Agent (agents Self))
    (display Agent)))


(defmethod DISPLAY :after ((Self agent-3d))
  ;; Conclude transformations
  (glPopmatrix))


(defmethod ANIMATE ((Self agent-3d) Time)
  ;; animate attached agents
  (dolist (Agent (agents Self))
    (animate Agent Time))
  ;; velocity
  (incf (x Self) (* Time (velocity-x Self)))
  (incf (y Self) (* Time (velocity-y Self)))
  (incf (z Self) (* Time (velocity-z Self)))
  ;; acceleration
  (incf (velocity-x Self) (* Time (acceleration-x Self)))
  (incf (velocity-y Self) (* Time (acceleration-y Self)))
  (incf (velocity-z Self) (* Time (acceleration-z Self))))


(defmethod ATTACH ((Self agent-3d) Agent)
  ;; attach subgent to this agent
  (push Agent (agents Self)))


(defmethod KEY-EVENT-HANDLER ((Self agent-3d) Key)
  ;; pass key events to subagents
  (dolist (Agent (agents Self))
    (key-event-handler Agent Key)))

;_______________________________________
;  Sphere                               |
;_______________________________________

(defclass SPHERE (agent-3d)
  ((size :accessor size :initform 5.0 :initarg :size)
   (quadric :accessor quadric :initform (gluNewQuadric))))


(defmethod INITIALIZE-INSTANCE ((Self sphere) &rest Initargs)
  (declare (ignore Initargs))
  (call-next-method)
  (gluQuadricDrawstyle (quadric Self) glu_fill)
  (gluQuadricOrientation (quadric Self) glu_outside)
  (gluQuadricNormals (quadric Self) glu_smooth)
  (gluQuadricTexture (quadric Self) gl_true))
  

(defmethod DISPLAY ((Self sphere))
  (glEnable gl_lighting)
  (glTexenvf gl_texture_env gl_texture_env_mode gl_replace)
  (gluSphere (quadric Self) (size Self) 20 20)
  (call-next-method))

;_______________________________________
;  Sky Dome                             |
;_______________________________________

(defclass SKY-DOME (sphere)
  ())


(defmethod INITIALIZE-INSTANCE ((Self sky-dome) &rest Initargs)
  (declare (ignore Initargs))
  ;; sky domes must be very LARGE
  (setf (size Self) 100.0)
  (call-next-method))


(defmethod DISPLAY ((Self sky-dome))
  (glDisable gl_lighting)
  (glEnable gl_texture_2d)
  (glTexenvf gl_texture_env gl_texture_env_mode gl_decal)
  (gluQuadricOrientation (quadric Self) glu_inside)
  (use-texture (view Self) "skyIsland512x256.png")
  (call-next-method))


#| Examples:


(defclass SILLY-GAME-WINDOW (game-window) ())

(defmethod COMPOSE-SCENE ((Self silly-game-window))
  (call-next-method)
  (attach Self (make-instance 'cursor-controled-sphere)))


(defclass CURSOR-CONTROLED-SPHERE (sphere) ())

(defmethod KEY-EVENT-HANDLER ((Self cursor-controled-sphere) Key)
  (let ((Delta 0.1s0))
    (case Key
      (#\ForwardArrow
       (setf (velocity-x Self) 0s0)
       (incf (acceleration-x Self) Delta))
      (#\BackArrow
       (setf (velocity-x Self) 0s0)
       (decf (acceleration-x Self) Delta))
      (#\UpArrow (decf (velocity-z Self) Delta))
      (#\DownArrow (incf (velocity-z Self) Delta)))))

  


(defparameter *Window* (make-instance 'silly-game-window))

(start-animation *Window*)

(compose-scene *Window*)

(aim-camera (camera *Window*) :eye-z 40.0 :eye-x 0.0 :eye-y 0.0 :near 1.0)

(frame-rate *Window*)


|#