;;;-*- Mode: Lisp; Package: ccl -*-
;*********************************************************************
;*                                                                   *
;*        O p e n G L  for  M C L                                    *
;*                                                                   *
;*********************************************************************
   ;* Author    : Alexander Repenning (alexander@agentsheets.com)    *
   ;*             http://www.agentsheets.com                         *
   ;* Copyright : (c) 1996-2002, AgentSheets Inc.                    *
   ;* Filename  : OpenGL for MCL.lisp                                *
   ;* Updated   : 12/11/02                                           *
   ;* Version   :                                                    *
   ;*   0.0     : 08/28/00 on the way to Switzerland                 *
   ;*   0.1     : 09/18/00 opengl-window                             *
   ;*   0.1.1   : 09/21/00 avoid null ptr on init                    *
   ;*   0.1.2   : 09/23/00 Text functions, don't crash MCL on quit   *
   ;*   0.2     : 03/24/01 Error handling                            *
   ;*   0.3     : 04/14/01 Cameras                                   *
   ;*   0.4     : 07/10/01 Vector functions                          *
   ;*   0.5     : 08/14/01 Full Screen support                       *
   ;*   0.5.1   : 12/26/01 Live Resize                               *
   ;*   0.6     : 01/05/02 OS X carbon/Aqua                          *
   ;*   0.7     : 02/19/02 32 bit RGBA Texture File Interface        *
   ;*   0.7.1   : 03/06/02 added 24 bit RGB                          *
   ;*   0.7.2   : 03/10/02 carbonized create-texture-from-file       *
   ;*   0.7.3   : 03/13/02 antialiased textures                      *
   ;*   0.7.4   : 05/23/02 G4 radeon 7500: no more agl_accelerated   *
   ;*   0.7.5   : 06/26/02 32 bit mode after full screen esc         *
   ;*   0.8     : 11/06/02 use-texture texture manager               *
   ;*   0.8.1   : 11/10/02 vertical texture orientation fixed        *
   ;*   0.8.2   : 11/27/02 watch curson on texture load              *
   ;*   1.0     : 11/29/02 OS X carbon only                          *
   ;*   1.1     : 12/11/02 ok to close animated windows, tex cleanup *
   ;* SW/HW     : PowerPC G4, MCL 5                                  *
   ;* Abstract  : OpenGL for Apple interface with MCL wrapper        *
   ;*                                                                *
   ;******************************************************************
#| Todos

- fix huge text memory leek caused by text textures (run example 4 with live resize ATI memory control panel) 
- texture cleaup when closing OpenGL view/window
|#

(in-package :ccl)

(use-package :gl)

;********************************************
;*  Class: OPENGL-Simple-View               *
;********************************************

(defclass OPENGL-SIMPLE-VIEW (simple-view)
  ((aglcontext :accessor aglcontext :initform nil :documentation "current glContext")
   (current-font :accessor current-font :initform nil)
   (camera :accessor camera)
   (textures :accessor textures :initform (make-hash-table :test #'equal))
   (codebase :accessor codebase :initform (and *Loading-File-Source-File* (directory-namestring *Loading-File-Source-File*)))
   (animation-time :accessor animation-time :initform (let ((Time (make-record :unsignedwide))) (#_MicroSeconds Time) Time))
   (animation-process :accessor animation-process :initform nil)
   (animation-name-index :accessor animation-name-index :initform 0 :allocation :class))
  (:documentation "Super class for OpenGL views, windows, and full screens"))


(defmethod MAKE-AGL-CONTEXT ((Self opengl-simple-view))
  (let ((attributes {AGL_RGBA
                    AGL_DOUBLEBUFFER
                    AGL_DEPTH_SIZE 32
                    AGL_NONE}))
    ;; create context
    (rlet ((GDhandleArray :pointer))
      (%put-ptr GDhandleArray (#_getMainDevice))
      (let ((AGLPixelFormat (aglChoosePixelFormat GDhandleArray 1 attributes)))
        (agl-error-check "pixel format")
        (#_disposePtr attributes)
        (setf (aglcontext Self) (aglCreateContext AGLPixelFormat (%null-ptr)))
        ;; SHOULD DO share contexts to share display lists and textures
        ;; texture sharing only works when FIRST creating all contexts and THEN creating textures:
        ;; http://developer.apple.com/qa/qa2001/qa1031.html
        (agl-error-check "context")
        (aglDestroyPixelFormat AGLPixelFormat)
        (agl-error-check)))))


(defmethod INITIALIZE-INSTANCE ((Self opengl-simple-view) &rest Initargs)
  (apply #'call-next-method Self 
         :view-size (or (getf Initargs :view-size) #@(100 100))
         :view-position (or (getf Initargs :view-position) #@(50 50))
         Initargs)
  (make-agl-context Self))


(defmethod CLEANUP-AGL ((Self opengl-simple-view)) "
  Make sure to call this method only if the aglContext is no longer used"
  (maphash 
   #'(lambda (Key Value) 
       (declare (ignore Value))
       (rlet ((&Name :GLint))
         (%put-word &Name Key)
         (glDeleteTextures 1 &Name)))
   (textures Self))
  (aglSetDrawable (aglcontext Self) (%null-ptr))
  (agl-error-check)
  (aglSetCurrentContext (%null-ptr))
  (agl-error-check)
  (aglDestroyContext (aglcontext Self))
  (agl-error-check))

;;___________________________________
;; high level (GLUT-like) methods    |
;;__________________________________/

(defgeneric DISPLAY (opengl-simple-view)
  (:documentation "Draw content of OpenGL window"))

(defgeneric RESHAPE (opengl-simple-view Width Height)
  (:documentation "Adjust transformations in case of window geometry change"))

(defgeneric IDLE (opengl-simple-view)
  (:documentation "This method is called if window is active and OpenGL has spare CPU cycle. Typical application: simple animation."))

(defgeneric INIT (opengl-simple-view)
  (:documentation "Chance to initialize OpenGL context"))

(defgeneric ANIMATE (opengl-simple-view Time)
  (:documentation "Update content but do not draw/display. Time is in micro seconds. Use time to implement frame rate independent animation"))

;;___________________________________________________
;; default implementations of high level functions   |
;;__________________________________________________/

(defmethod DISPLAY ((Self opengl-simple-view))
  ;; default drawing: the obligatory RGB OpenL triangle
  (glClearColor 0.0 0.0 0.0 1.0)        ; black looks good
  (glClear (logior GL_COLOR_BUFFER_BIT gl_depth_buffer_bit))
  (glShadeModel gl_smooth)
  ;; use complete color spectrum
  (glBegin gl_polygon)
    (glColor3f 1.0 0.0 0.0)
    (glVertex3f 0.0 1.0 0.0) 
    (glColor3f 0.0 1.0 0.0)
    (glVertex3f -1.0 -1.0 0.0) 
    (glColor3f 0.0 0.0 1.0)
    (glVertex3f 1.0 -1.0 0.0) 
  (glEnd))


(defmethod RESHAPE ((Self opengl-simple-view) Width Height)
  (aglSetCurrentContext (aglcontext Self))
  (agl-error-check)
  (glviewport 0 0 Width Height)
  (unless (aglUpdateContext (aglcontext Self))
    (agl-error-check)
    (error "OpenGL could not adjust viewport to new size"))
  (aim-camera (camera Self) :aspect (float (/ Width Height))))


(defmethod IDLE ((Self opengl-simple-view)))


(defmethod INIT ((Self opengl-simple-view)))

;;___________________________________________________
;; Animation and timer                               |
;;__________________________________________________/


(defmethod DELTA-TIME ((Self opengl-simple-view)) "
  Return time passed since last animation."
  (rlet ((Time :unsignedwide))
    (#_MicroSeconds Time)
    (#_WideSubtract Time (animation-time Self))
    (#_MicroSeconds (animation-time Self))
    ;; An unsigned wide "only" holds the delta time for t < 1 hour! 
    ;; unless if the animation is less than 1 frame / hour this will not be an issue
    (/ (rref Time :UnsignedWide.lo) 1000000.0)))


(defmethod ANIMATE ((Self opengl-simple-view) Time)
  (declare (ignore Time)))


(defmethod START-ANIMATION ((Self opengl-simple-view))
  (cond
   ;; don't restart an existing, active process
   ((and (animation-process Self)  
         ;; a non active process could be the result of a initiated user process abort
         (process-active-p (animation-process Self)))  
    (warn "animation is already running"))
   (t  ;; create a new process
    (incf (animation-name-index Self))
    (setf (animation-process Self)
          (process-run-function
           (format nil "OpenGL Animation: ~A~A" 
                   (string-capitalize (string (type-of Self)))
                   (animation-name-index Self))
           #'(lambda ()
               (loop
                 (without-interrupts
                  (animate Self (delta-time Self))
                  (view-draw-contents Self)))))))))


(defmethod STOP-ANIMATION ((Self opengl-simple-view))
  (when (animation-process Self)
    (process-abort (animation-process Self))
    (setf (animation-process Self) nil)))

;;___________________________________________________
;; Tools                                             |
;;__________________________________________________/

(defmethod FRAME-RATE ((Self opengl-simple-view))
  ;; cycle through frames from one second and count the number of frames drawn
  (aglSetCurrentContext (aglcontext Self))
  (agl-error-check)
  (without-interrupts   ;; stop other animations etc.
   (let ((Frame-Count 0)
         (Stop-Time (+ (get-internal-run-time) (* 1.0 internal-time-units-per-second))))
     (loop
       (display Self)
       (aglSwapbuffers (aglcontext Self))
       (agl-error-check)
        (incf Frame-Count)
       (when (>= (get-internal-run-time) Stop-Time)
         (return Frame-Count))))))

;_________________________
;; MCL event handlers     |
;________________________/

(defmethod VIEW-DRAW-CONTENTS ((Self opengl-simple-view))
  ;; draw contents
  (aglSetCurrentContext (aglcontext Self))
  (agl-error-check)
  (display Self)
  (aglSwapbuffers (aglcontext Self))
  (agl-error-check))


(defmethod VIEW-CLICK-EVENT-HANDLER ((Self opengl-simple-view) Where)
  (declare (ignore Where))
  (call-next-method)
  (aglSetCurrentContext (aglcontext Self))
  (agl-error-check)
  (track-mouse-3d (camera Self)))


(defmethod VIEW-KEY-EVENT-HANDLER ((Self opengl-simple-view) Char)
  (declare (ignore Char))
  (ed-beep))

;;____________________
;; Textures           |
;;____________________/

(defmethod TEXTURE-DIRECTORY ((Self opengl-simple-view))
  ;; return the directory which should contain all the texture files
  (concatenate 'string (codebase Self) "data:" "textures:"))


(defmethod TEXTURE-FILE ((Self opengl-simple-view) Texture-Name)
  (let* ((File (make-pathname :directory (texture-directory Self) :name Texture-Name))
         (True-File (probe-file File)))
    (unless True-File (error (format nil "Missing texture file: ~A" File)))
    True-File))


(defmethod USE-TEXTURE ((Self opengl-simple-view) Texture-Name)
  (let ((Texture-Id (gethash Texture-Name (textures Self))))
    ;; load texture from file if necessary
    (unless Texture-Id
      (with-cursor *Watch-Cursor*
        (setq Texture-Id (create-texture-from-file (texture-file Self Texture-Name))))
      (setf (gethash Texture-Name (textures Self)) Texture-Id))
    ;; make the texture active (for now assume a 2D texture
    (glbindtexture gl_texture_2d Texture-Id)))

;;____________________
;; Text functions     |
;;____________________/

(defmethod MAKE-OPENGL-FONT ((Self opengl-simple-view) &optional (Font-Id $geneva) (Font-Size 24) (Font-Style 0))
  ;; make an OpenGL font and make it the current font
  (let ((Font (glgenlists 96)))
    (aglUseFont (aglcontext Self) 
                Font-Id 
                Font-Style   ;; bitset: {bold, italic, underline, outline, shadow, condense, extend}
                Font-Size  
                32  ;; starting at the space
                96  ;; number of characters in set
                Font)
    (agl-error-check)
    (setf (current-font Self) Font)
    Font))


(defmethod GLPRINT ((Self opengl-simple-view) String &optional Font)
  (let ((Font (or Font (current-font Self) (make-opengl-font Self))))
    (glPushAttrib gl_list_bit)
    (glListBase (- Font 32))
    (with-cstr (&String String)
      (glCallLists (length String) gl_unsigned_byte &String))
    (glPopAttrib)))


;;***********************************
;; OPENGL-View                      *
;;***********************************

(defclass OPENGL-VIEW (opengl-simple-view)
  ((window-height :accessor window-height :initform nil :documentation "OpenGL views need to track the height of their containing windows"))
  (:documentation "an OpenGL view to be put into windows joining other views and dialog items"))


(defmethod ADJUST-BUFFER-RECT ((Self opengl-view))
  (let ((&rect {0 0 0 0}))
    (%put-long &rect (point-h (view-position Self)) 0)
    (%put-long &rect (- (point-v (view-size (view-window Self)))
                        (point-v (view-position Self))
                        (point-v (view-size Self))) 4)
    (%put-long &rect (point-h (view-size Self)) 8)
    (%put-long &rect (point-v (view-size Self)) 12)
    (aglSetInteger (aglcontext Self) AGL_BUFFER_RECT &rect)
    (agl-error-check)
    (#_DisposePtr &rect)))


(defmethod INSTALL-VIEW-IN-WINDOW ((Self opengl-view) Window)
  (call-next-method)
  ;; Set Drawable could not be called earlier since the view must be contained in an existing window
  (aglSetDrawable (aglcontext Self) (#_GetWindowPort (wptr Window)))
  (agl-error-check)
  (aglSetCurrentContext (aglcontext Self))
  (agl-error-check)
  ;; adjust the buffer rect to view
  (setf (window-height Self) (point-v (view-size Window)))
  (aglEnable (aglcontext Self) AGL_BUFFER_RECT)
  (adjust-buffer-rect Self)
  (glViewport 0 0 (point-h (view-size Self)) (point-v (view-size Self)))
  (glclear gl_color_buffer_bit)
  (glfinish)
  (setf (camera Self) (make-instance (camera-class self) :view Self))
  (aim-camera (camera Self))
  (init Self))

;by AAA 15/09/03

(defmethod camera-class ((Self t)) 'camera)


(defmethod REMOVE-VIEW-FROM-WINDOW ((Self opengl-view))
  (call-next-method)
  (cleanup-agl Self))

;_________________________
;; MCL event handlers     |
;________________________/

(defmethod VIEW-DRAW-CONTENTS ((Self opengl-view))
  ;; adjust position if window height changed
  (unless (equal (window-height Self) (point-v (view-size (view-window Self))))
    (adjust-buffer-rect Self)
    (setf (window-height Self) (point-v (view-size (view-window Self)))))
  (call-next-method))

;;*******************************
;; OPENGL-Window                *
;;*******************************

(defclass OPENGL-WINDOW (window opengl-simple-view)
  ((full-screen-p :accessor full-screen-p :initform nil :initarg :full-screen-p
                       :documentation "If non-nil this view operates as full screen but still behaves like a window"))
  (:documentation "Basic OpenGL 3D graphics viewer window"))


(defmethod MAKE-AGL-CONTEXT ((Self opengl-window))
  (if (full-screen-p Self)
    ;; pixel format is full screen specific
    (let ((attributes {AGL_RGBA
                      AGL_DOUBLEBUFFER
                      AGL_FULLSCREEN    
                      AGL_DEPTH_SIZE 32
                      AGL_NONE}))
      ;; create context
      (rlet ((GDhandleArray :pointer))
        (%put-ptr GDhandleArray (#_getMainDevice))
        (let ((AGLPixelFormat (aglChoosePixelFormat GDhandleArray 1 attributes)))
          (agl-error-check "after pixel format")
          (#_disposePtr attributes)
          (setf (aglcontext Self) (aglCreateContext AGLPixelFormat (%null-ptr)))
          ;; SHOULD DO share contexts to share display lists and textures
          ;; texture sharing only works when FIRST creating all contexts and THEN creating textures:
          ;; http://developer.apple.com/qa/qa2001/qa1031.html
          (agl-error-check "context")
          (aglDestroyPixelFormat AGLPixelFormat)
          (agl-error-check))))
    (call-next-method)))


(defmethod END-FULL-SCREEN-MODE ((Self opengl-window))
  (when (full-screen-p Self)
    (aglSetDrawable (aglcontext Self) (%null-ptr))))


(defmethod BEGIN-FULL-SCREEN-MODE ((Self opengl-window) &optional Width Height)
  (aglSetCurrentContext (aglcontext Self))
  (agl-error-check "set Context")
  ;; default full screen size to current size of main screen
  (when (or (not Width) (not Height))
    (let ((Size (rref (#_getMainDevice) :gdevice.gdRect.botright)))
      (setf Width (point-h Size))
      (setf Height (point-v Size))))
  (without-interrupts
  (unwind-protect
    (progn
      (aglSetFullScreen (aglcontext Self) Width Height 85 0)
      (glClear gl_color_buffer_bit)
      (glFinish)
      (glViewport 0 0 Width Height)
      (aim-camera (camera Self) :aspect (float (/ Width Height)))
      (init Self)
      (view-draw-contents Self)
      ;; dispatch until users teminates
      (full-screen-event-dispatch Self))
    (end-full-screen-mode Self))))


(defmethod INITIALIZE-INSTANCE ((Self opengl-window) &rest Initargs)
  (apply #'call-next-method Self 
         :color-p t 
         :window-show nil
         :view-size (or (getf Initargs :view-size) #@(300 300))
         :window-title (or (getf Initargs :window-title)
                           (format nil "OpenGL: ~A" (string-capitalize
                                                     (string (type-of Self)))))
         Initargs)
  (make-agl-context Self)
  (cond
   ;; Full Screen
   ((full-screen-p Self)
    (setf (camera Self) (make-instance 'camera :view Self)))
   (t     ;; Window
    ;; the entire window content is one glContext
    (aglSetDrawable (aglcontext Self) (#_GetWindowPort (wptr Self)))
    (agl-error-check)
    (aglSetCurrentContext (aglcontext Self))
    (glclear gl_color_buffer_bit)
    (glfinish)
    (setf (camera Self) (make-instance 'camera :view Self))
    (aim-camera (camera Self))
    (init Self)
    (window-show Self))))


(defmethod FULL-SCREEN-EVENT-DISPATCH ((Self opengl-window))
  ;; This is a tigth version of the general event dispatcher
  ;; only used in full screen mode
  ;; No other events will be dispatched
  (rlet ((Event :EventRecord))
    (without-interrupts
     (loop
       ;; Input events
       (when (event-available-p)
         (get-next-event Event)
         (case (rref Event :eventRecord.what)
           (#.#$mouseDown
            (view-click-event-handler Self (rref Event :eventRecord.where)))
           ((#.#$keyDown #.#$autoKey)
            (let ((Char (code-char (logand #$charCodeMask (rref Event :eventRecord.message)))))
              (case Char
                (#\ESC  ;; use the #\ESC key to escape from full screen mode
                 (return))
                (#\f 
                 (format t "~%frame rate: ~A fps" (frame-rate Self)))
                (t
                 (view-key-event-handler Self Char)))))))
       ;; Animation
       (animate Self (delta-time Self))
       (view-draw-contents Self)))))


(defmethod START-ANIMATION ((Self opengl-window))
  ;; if view is FULL SCREEN then animation is taken care of in event dispatch
  (when (not (full-screen-p Self))
    (call-next-method)))

;_________________________
;; MCL event handlers     |
;________________________/

(defmethod VIEW-DRAW-CONTENTS ((Self opengl-window))
  (aglSetCurrentContext (aglcontext Self))
  (agl-error-check)
  (display Self)
  (aglSwapbuffers (aglcontext Self))
  (agl-error-check))

;; These handlers do NOT get called ever when view is FULL SCREEN

(defmethod WINDOW-ZOOM-EVENT-HANDLER ((Self opengl-window) Message)
  (declare (ignore Message))
  (call-next-method)
  (reshape Self (point-h (view-size Self)) (point-v (view-size Self)))
  (invalidate-view Self))


(defmethod WINDOW-GROW-EVENT-HANDLER ((Self opengl-window) Where)
  (cond
   ;; OS 9 style resize
   ((command-key-p) 
    (call-next-method)
    (reshape Self (point-h (view-size Self)) (point-v (view-size Self)))
    (invalidate-view Self))
   ;;  OS X Live Resize
   (t   
    (let ((Delta-Size (add-points (view-position Self) (subtract-points (view-size Self) Where)))
          (Last-Position nil))
      (with-focused-view Self
        (loop
          ;; sustain animation
          (animate Self (delta-time Self))
          (view-draw-contents Self)
          ;; resize
          (when (not (mouse-down-p)) (return))
          (let ((Position (view-mouse-position Self)))
            (when (and (shift-key-p) Last-Position)         ;; use shift key as modifier to constrain vertical size
              (setq Position (make-point (point-h Position) (point-v Last-Position))))
            (when (not (equal Position Last-Position))
              (setq Last-Position Position)
              (let ((New-Size (add-points Position Delta-Size)))
                (set-view-size Self (point-h New-Size) (point-v New-Size))
                 (reshape Self (point-h (view-size Self)) (point-v (view-size Self)))
                 (view-draw-contents Self)
                 (#_QDFlushPortBuffer (#_GetWindowPort (wptr Self)) (%null-ptr)))  ))))))))


(defmethod WINDOW-DRAG-EVENT-HANDLER ((Self opengl-window) Where)
  ;;  OS X Live Drag
  (declare (ignore Where))
  (let ((Offset (subtract-points (view-position Self) (view-mouse-position nil))))
    (with-focused-view Self
      (loop
        (when (not (mouse-down-p)) (return))
        ;; drag
        (set-view-position Self (add-points Offset (view-mouse-position nil)))
        ;; sustain animation
        (animate Self (delta-time Self))
        (view-draw-contents Self)))))


(defmethod WINDOW-NULL-EVENT-HANDLER ((Self opengl-window))
  (call-next-method)
  (aglSetCurrentContext (aglcontext Self))
  (agl-error-check)
  (idle Self))


(defmethod WINDOW-CLOSE-EVENT-HANDLER ((Self opengl-window))
  (stop-animation Self)
  (call-next-method))


(defmethod WINDOW-CLOSE ((Self opengl-window))
  (call-next-method)
  (cleanup-agl Self))


#| Example: Window mode versus Full Screen mode

(defparameter *Window* (make-instance 'opengl-window))

(frame-rate *Window*)



(defparameter *Screen* (make-instance 'opengl-window :full-screen-p t))


(begin-full-screen-mode *Screen*)

(begin-full-screen-mode *Screen* 640 480)

|#


#| Example: Make a Window containing 4 opengl-views


(defparameter *W* (make-instance 'window :color-p t))


(add-subviews *W*
  (make-instance 'opengl-view :view-position #@(10 10))
  (make-instance 'opengl-view :view-position #@(120 10))
  (make-instance 'opengl-view :view-position #@(10 120))
  (make-instance 'opengl-view :view-position #@(120 120)))

|#


#| Example 1: Instance of OpenGL-Window class

(defparameter *Window* (make-instance 'opengl-window))


(frame-rate *Window*)

|#


#| Example 2: Camera motion

(defclass TRIANGLE-FLIGHT-WINDOW (opengl-window)
  ())


(defmethod DISPLAY ((Self triangle-flight-window))
  (call-next-method)
  (dotimes (i 10)
    (let ((z (* i 0.5)))
      (glBegin gl_polygon)
      (glColor3f 1.0 0.0 0.0)
      (glVertex3f 0.0 1.0 z) 
      (glColor3f 0.0 1.0 0.0)
      (glVertex3f -1.0 -1.0 z) 
      (glColor3f 0.0 0.0 1.0)
      (glVertex3f 1.0 -1.0 z) 
      (glEnd))))


(defmethod INIT ((Self triangle-flight-window))
  (call-next-method)
  (glenable gl_depth_test)
  (aim-camera (camera Self) :eye-z 10.0))


(defparameter *Window* (make-instance 'triangle-flight-window))


|#



#| Example 3: Cube with shading


(defclass CUBE-OPENGL-WINDOW (opengl-window) ())

(defmethod INIT ((Self cube-opengl-window))
  (glClearColor 0.5 0.5 0.5 1.0)
  (glShadeModel gl_smooth)
  ;; define material
  (with-rgba-vector Specular (0.5 0.5 0.5 0.0)
    (glmaterialfv gl_front_and_back gl_specular Specular))
  (glmaterialf gl_front_and_back gl_shininess 128.0)
  ;; light
  (with-rgba-vector Position (1.0 1.0 1.0 1.0)
    (gllightfv gl_light0 gl_position Position))
  (with-rgba-vector White-Light (1.0 1.0 1.0 1.0)
    (gllightfv gl_light0 gl_diffuse White-Light))
  (with-rgba-vector Specular-Light (0.5 0.5 0.5 1.0)
    (gllightfv gl_light0 gl_specular Specular-Light))
  ;; enablers
  (glenable gl_lighting)
  (glenable gl_light0)
  (glenable gl_depth_test)
  (aim-camera (camera Self) :eye-z 2.0))


(defmethod DISPLAY ((Self cube-opengl-window))
 (glClear (logior GL_COLOR_BUFFER_BIT gl_depth_buffer_bit))
 (let* ((A 0.6)
        (-A (- A)))
   (with-rgba-vector V (1.0 0.0 0.0 1.0)
     (glmaterialfv gl_front_and_back gl_ambient_and_diffuse V))
   (glBegin GL_POLYGON)
     (glnormal3f 0.0 0.0 1.0)
     (glVertex3f -a -a a)
     (glVertex3f -a  a a)
     (glColor4f 0.0 1.0 0.0 1.0)
     (glVertex3f  a  a a)
     (glColor4f 0.0 0.0 1.0 1.0)
     (glVertex3f  a -a a)
   (glEnd)
   (with-rgba-vector V (0.0 1.0 0.0 1.0)
     (glmaterialfv gl_front_and_back gl_ambient_and_diffuse V))
   (glBegin GL_POLYGON)
     (glnormal3f 0.0 0.0 1.0)
     (glVertex3f -a -a -a)
     (glVertex3f -a  a -a)
     (glVertex3f  a  a -a)
     (glVertex3f  a -a -a)
   (glEnd)
   (with-rgba-vector V (0.0 0.0 1.0 1.0)
     (glmaterialfv gl_front_and_back gl_ambient_and_diffuse V))
   (glBegin GL_POLYGON)
     (glnormal3f -1.0 0.0 0.0)
     (glVertex3f -a -a -a)
     (glVertex3f -a -a  a)
     (glVertex3f -a  a  a)
     (glVertex3f -a  a -a)
   (glEnd)
   (with-rgba-vector V (0.0 1.0 1.0 1.0)
     (glmaterialfv gl_front_and_back gl_ambient_and_diffuse V))
   (glBegin GL_POLYGON)
     (glnormal3f 1.0 0.0 0.0)
     (glVertex3f  a -a -a)
     (glVertex3f  a -a  a)
     (glVertex3f  a  a  a)
     (glVertex3f  a  a -a)
   (glEnd)))

(defparameter *Cube-Window* (make-instance 'cube-opengl-window :full-screen-p t))

(begin-full-screen-mode *Cube-Window* 640 480)


|#

#|
;; Example 4: Plot

(defclass 3D-PLOT-WINDOW (opengl-window)
  ((data-vector :accessor data-vector :initform (make-array 200 :initial-element 0.0))
   (data-offset :accessor data-offset :initform 0.0)))


(defmethod GENERATE-SIN-DATA-VECTOR ((Self 3d-plot-window))
  (dotimes (i (array-dimension (data-vector Self) 0))
    (setf (svref (data-vector Self) i) (* 0.4 (sin (+ (data-offset Self) (* i 0.09)))))))


(defmethod INIT ((Self 3d-plot-window))
  (generate-sin-data-vector Self)
  (glclearcolor 0.0 0.0 0.0 1.0)
  (glShadeModel gl_smooth)
  ;; define material
  (with-rgba-vector Specularity (0.5 0.5 0.5 1.0)
    (glmaterialfv gl_front_and_back gl_specular Specularity))
  (glmaterialf gl_front_and_back gl_shininess 10.0)
  ;; light 0
  (with-rgba-vector Position (1.0 1.0 1.0 1.0)
    (gllightfv gl_light0 gl_position Position))
  (with-rgba-vector Light (1.0 1.0 1.0 1.0)
    (gllightfv gl_light0 gl_diffuse Light))
  (with-rgba-vector Light (0.5 0.5 0.5 1.0)
    (gllightfv gl_light0 gl_specular Light))
  ;; enablers
  (glenable gl_lighting)
  (glenable gl_light0)
  (glenable gl_depth_test))


(defmethod DISPLAY ((Self 3d-plot-window)) 
  (glClear (logior GL_COLOR_BUFFER_BIT gl_depth_buffer_bit))
  (glenable gl_normalize)
  (with-rgba-vector V (0.7 0.0 0.0 0.5)
    (glmaterialfv gl_front_and_back gl_ambient_and_diffuse V))
  (let* ((Vector (data-vector Self))
         (x -1.0) 
         (dx (/ 2.0 (array-dimension Vector 0)))
         (dz 0.8))
    ;; draw the data segments
    (dotimes (i (1- (array-dimension Vector 0)))
      (let ((dy (- (aref Vector (1+ i)) (aref Vector i))))
        (glBegin GL_POLYGON)
        (compute-normal dx 0.0 0.0 0.0 dy 0.0 0.0 0.0 dz)       ;; because of 0.0s could be done more efficiently
        (glVertex3f  x (aref Vector i) 0.0)
        (glVertex3f (+ x dx) (aref Vector (1+ i)) 0.0)
        (glVertex3f (+ x dx) (aref Vector (1+ i)) dz)
        (glVertex3f  x (aref Vector i) dz))
      (glEnd)
      (incf x dx))))
    
(make-instance '3d-plot-window)


|#

#| Example 5: Text

(defclass TEXT-OPENGL-WINDOW (opengl-window)
  ((small-font :accessor small-font)
   (large-font :accessor large-font)
   (giant-font :accessor giant-font)))
  

(defmethod INIT ((Self text-opengl-window))
  (setf (small-font Self) (make-opengl-font Self $geneva 96))
  (setf (large-font Self) (make-opengl-font Self $times 256))
  (setf (giant-font Self) (make-opengl-font Self $times 350)))


(defmethod DISPLAY ((Self text-opengl-window))
  (glClear (logior gl_color_buffer_bit gl_depth_buffer_bit))
  (glclearcolor 0.0 0.0 0.5 1.0)
  (glColor3f 0.5 0.5 0.5)
  (glrasterpos2f -0.1 0.2)
  (glprint Self "Where" (small-font Self))
  (glColor3f 0.5 0.1 0.1)
  (Glrasterpos2F -0.1 0.5)
  (glprint Self "am I" (large-font Self))
  (glColor3f 0.8 0.8 0.8)
  (Glrasterpos2F -0.1 0.7)
  (glprint Self "now?" (giant-font Self)))



(make-instance 'text-opengl-window)

|#