;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon

(in-package :om)
;==========================================================================
; OpenMusic : ScoreOMInterface.lisp
;==========================================================================


;--------------------
;  CONCAT
;--------------------

(defmethod* concat ((s1 sequence*) (s2 sequence*))
  :initvals (list (make-instance 'voice) (make-instance 'voice)) 
  :indoc '("a music sequence" "a music sequence")
  :icon 230
  :doc "Concatenates 2 music sequences into a new one.

 Voice x Voice  --> Voice
 Measure x Voice --> Voice
 Voice x Measure --> Voice
 multi-seq x multi-seq --> multi-seq
 chord-seq x chord-seq --> chord-seq
 poly x poly --> poly
" 
  (let ((frac-min (fraction-minimale-commune s1 s2))
        (ss1 (duplique-structure-musicale s1))
        (ss2 (duplique-structure-musicale s2)))
    (change-qvalue ss1 frac-min)
    (change-qvalue ss2 frac-min)
    (loop for item in (inside ss2) do (setf (offset item) (+ (offset item) (extent ss1)))) 
    (mki (type-of s1) 
             :empty t
             :offset 0
             :extent (+ (extent ss1) (extent ss2))
             :qvalue frac-min
             :inside (append (inside ss1) (inside ss2))
          ) ))

(defmethod* concat ((s1 measure) (s2 voice))
  (concat (measure->voice s1) s2))

(defmethod* concat ((s1 voice ) (s2 measure))
  (concat  s1 (measure->voice s2)))


(defmethod* concat ((s1 poly ) (s2 poly))
  (mki 'poly
       :voices (mapcar 'concat (inside s1) (inside s2)) ))

(defmethod* concat ((s1 multi-seq ) (s2 multi-seq))
  (mki 'multi-seq 
       :chord-seqs
       (loop 
         with chs1 = (inside s1)
         with chs2 = (inside s2)
         while (or chs1 chs2)
         collect (if chs1
                   (if chs2 (concat (pop chs1) (pop chs2))
                       (pop chs1))
                   (pop chs2)))))



;--------------------
;  GET-CHORDS
;--------------------

(defmethod* get-chords ((self sequence*))
   :initvals (list (make-instance 'voice) ) 
   :indoc '("a music sequence")
   :icon 262
   :doc "Extract an ordered list (or list of list) of chords from a music sequence or superposition.
Time (Onset) information is removed.
Allowed input type  : voice, chord-seq, poly, multi-seq.
"
   (loop for sub in (inside self)
         when (not (cont-chord-p sub))
         if (note-or-chord-p sub) collect (ot-clone sub)
         else if (container-p sub) append (chords sub)))


(defmethod* get-chords ((self superposition))
  (mapcar 'get-chords (inside self)))

(defmethod* get-chords ((self chord))
  (clone self))



;--------------------
;  MASK
;--------------------


(defmethod* mask ((obj1 sequence*) (obj2 sequence*) &optional (mode 1))
   :initvals (list (make-instance 'voice) (make-instance 'voice) 1) :indoc '("A voice" "A voice" "1 or -1")
   :doc "Takes 2 voices. Masks the 1st one by the 2nd one, that is play the 1st when the 2nd is not playing. 
Optional argument <mode>  is 1 or -1. In the 2nd case, the mask is inverted." 
   :icon 263
   (normalize-chord obj1)
   (normalize-chord obj2)
   (om-masque obj1 obj2 :mode (if (= mode 1) 'masque 'demasque)))


;--------------------
;  MERGER
;--------------------



;===================temporary merge correction===================================
;by KA 06-11-2001
(defun correct-merge (tree)
  
  (if (atom tree)
    tree
    (list (if (ratiop (first tree)) 
            (list (numerator (first tree)) 
                  (denominator (first tree))) 
            (first tree))
          (mapcar 'correct-merge (second tree)))))

(defmethod old-merger ((obj1 sequence*) (obj2 sequence*))
  (normalize-chord obj1)
  (normalize-chord obj2)
  (om-merge obj1 obj2))

(defmethod! merger ((self1 sequence*) (self2 sequence*)) 
   :initvals (list (make-instance 'voice) (make-instance 'voice)) 
   :indoc '("a voice or measure" "a voice or measure")
   :icon 253
   :doc "Merges rhytmically 2 voices or measures into a voice or measure"
   (let* ((result (old-merger self1 self2))
          (newself (clone result))
          (tree (tree newself))
          (newtree (list '? (cadr (correct-merge tree)))))
     (setf (tree newself) newtree)
     ))

;--------------------
;  SELECT
;--------------------

(defmethod* select ((self voice) (start integer) (end integer))
   :initvals (list (make-instance 'voice) 0 1) 
   :indoc '("a music sequence" "an integer" "an integer")
   :icon 330
   :doc "Extracts a subseqence from a music sequence.

WHEN :
<self> is a voice, <start> and <end> are measure numbers, result is a voice.
<self> is a poly, <start> and <end> are measure numbers, result is a poly
<self> is a chord-seq, <start> and <end> are absolute positions in ms, result is a chord-seq.
<self> is a multi-seq, <start> and <end> are absolute positions in ms, result is a multi-seq.
"
   
   (when (> start end) (Om-Message-abort "Extract-measures : Bad start/end parameters"))
   (let* ((measures (clone (subseq (inside self) 
                                   (max 0 (min start (1- (length (inside self)))))
                                   (min (1+ end) (length (inside self))))))
          (temp-voice (mki 'voice :empty t :inside measures))
          (first-chord (first-container temp-voice '(chord)))
          (last-chord (last-container temp-voice '(chord))))
     (when first-chord
       (when (cont-chord-p first-chord) (change-class first-chord 'chord))
       (loop for note in (inside first-chord)
             do (case (tie note)
                  (end (setf (tie note) nil))
                  (continue (setf (tie note) 'begin)))))
     (when last-chord
       (loop for note in (inside last-chord)
             do (case (tie note)
                  (begin (setf (tie note) nil))
                  (continue (setf (tie note) 'end)))))
     (loop for measure in measures
           with offset = (offset (first measures))
           do (setf (offset measure) (- (offset measure) offset)))
     (let ((voice (mki 'voice 
                       :empty t
                       :offset 0
                       :extent 0
                       :qvalue (qvalue self)
                       :inside measures
                       )))
       (setf (tempo voice)  (tempo self))
       (adjust-extent voice)
       voice)))


(defmethod* select ((self poly) (start integer) (end integer))
   (mki 'poly
        :voices (loop for voice in (inside self) 
                      collect (select voice start end))))

(defmethod* select ((self chord-seq) (start number) (end number))
   (unless (and (>= start 0)
                (< start end)
                (<= end (extent->ms self)))
     (Om-Message-abort "select : Bad start/end parameters"))
   (let ((chords 
          (loop for onset in (Lonset self)
                for chord in (inside self)
                when (and (>= onset start) (< onset end))
                collect (list (- onset start) chord))))
     (if chords
       (mki (type-of self) 
            :LMidic (mapcar 'second chords)
            :Lonset  (append (mapcar 'first chords) (list (- end start)))
            :legato 0)
       (mki (type-of self)
            :Lonset (list 0 (- end start))
            ))))

(defmethod* select ((self multi-seq) (start number) (end number))
   (mki 'multi-seq
        :chord-seqs
        (loop for chord-seq in (inside self)
              collect (select chord-seq start end))))



(defun load-obj-list-from-save (list)
  (loop for item in list
        collect (eval item)))

(defun get-real-chords (self)
  (loop for item in (collect-chords self)
        when (and (chord-p item) (not (cont-chord-p item))) collect item))

  

(defmethod execption-save-p ((self voice)) 'voice)
(defmethod save-exepcion ((self voice))
  `(when (find-class ',(type-of self) nil)
     (make-instance ',(type-of self)
       :tree ',(tree self)
       :chords (load-obj-list-from-save '(,.(mapcar #'(lambda (x) (omNG-save x)) (get-real-chords self))))
       :tempo ,(tempo self)
       :legato ,(legato self)
       :ties ',(ties self))))


