;******************************************************************
;-------| CHROMA SYSTEM
;-------| This file is: $LLsys/utils-cr.lisp
;-------| Implemented by Marco Stroppa
;-------| Version: 000817, updated 050221, 060814
;******************************************************************

; THIS FILE CONTAINS THE DEFINITION OF MIXED FUNCTIONS GENERALLY USEFUL EITHER
;   FOR THE CHROMA SYSTEM INTERNALLY ONLY OR BETWEEN CHROMA AND OM

(in-package cr)

;------------------------------------------------------------------
; FUNCTIONS (in alphabetical order):
;	fun->bpf
;	ran / ran-from / ran%
;	select
;------------------------------------------------------------------
; MACROS (in alphabetical order):
;	cassq
;	ifn
;	nextl
;	newl
;------------------------------------------------------------------
; METHODS! (in alphabetical order):
;	find-chroma-file
;------------------------------------------------------------------


;;;================================================================
;;;================================================================
;------------------------------------------------------------------
;;;; MACROS
; (cassq s l)
(defun cassq (sym l)
  (cdr (assoc sym l)))

; (ifn test body)
(defmacro ifn (testform elseform &body body)
  `(if (not ,testform) ,elseform (progn ,.body)))

; (nextl l [symb])
; (newl l [el])

(defmacro nextl (lst &optional symb)
  (if symb
    `(setq ,symb (pop ,lst))
    `(pop ,lst) ))

#|
(defun nextl (lst &optional symb)
  (if symb
    (setq symb (pop lst))
    (pop lst) ))
|#

(defmacro newl (lst elem) `(push ,elem ,lst))

;------------------------------------------------------------------c


;------------------------------------------------------------------
; fun->bpf

;	NAME:		fun->bpf  (CONSTRUCTOR)
;	TYPE:		Expr with 1 or 2 arguments
;	CALL:		(fun->bpf fun [precision])
;	FUNCTION:	return a bpf object of a given precision (default 0)
;	VALUE:		the bpf
;	SOURCE:		$LLsys/utils.lisp

(defun fun->bpf (fun &optional (precision 2))
"POLYMORPHIC FUNCTION:
   Convert a CHROMA FUN object into a BPF with optional precision.
   If a list of FUNs is given, return a list of BPFs with the same optional precision."
  (if (is_fun fun)
    (om::simple-bpf-from-list (x-list_fun fun) (y-list_fun fun) 'om::bpf precision)
    (fun-list->bpf fun precision)))

(defun fun-list->bpf (fun-list prec)
  (loop for f in fun-list
        collect (fun->bpf f prec)))

; (fun->bpf (make_fun '(0 0 1 1 .5 2)) 3)
; (fun->bpf (list (make_fun '(0 0 1 1 .5 2)) (make_fun '(0 0 1 1 .5 2 1 3))))



;------------------------------------------------------------------

;------------------------------------------------------------------
; select (rewrites a function that disappeared in OM 5)

;	NAME:		select  (SELCTOR)
;	TYPE:		Expr with 2 arguments
;	CALL:		(select <list> <pos>)
;	FUNCTION:	extracts the elements at position <pos> in <list> (1 = first element)
;	VALUE:		the list with the selected elements
;	SOURCE:		$LLsys/utils.lisp

(defun select (list pos)
"Extracts the elements at position <pos> from <list>.
 Contrary to OM's posn-match he first element is number 1."
   (om::posn-match list (om::om- pos 1)))
;(select '(1 2 3 4 5) '(2 3))


;------------------------------------------------------------------

; (ran [itvl] / [itvl var] )
; (ran% [itvl] / [itvl var] )
; (ran-from from to)
;	ran returns a random number comprised between -itvl and +itvl
;	   (-1/+1 if itvl is not present), or between itvl-var and itvl+var
;	ran% is the same as ran, but var is expressed in % [0-1] of itvl

(defun ran (&rest itvl)
  (let* ((a (car itvl))
         (b (cadr itvl))
         (from (ifn a
                    -1.0
                 (ifn b (* a -1) (- a b))))
         (to (ifn a
                  1.0
               (ifn b a (+ a b))))
         (size (abs (- to from)))
         (tmp to)
         )
    (when (> from to)		; SWAP ITVL IF NEEDED
      (setf to from)
      (setf from tmp) )
    (+ from
       (* size
          (/ (random 10000.0) 10000.0)))))

(defun ran-from (from to)
  (let ((itvl (- to from)))
    (+ from (random itvl))))


(defun ran% (&rest itvl)
    (let* ((a (car itvl))
	  (b (cadr itvl))
	  (from (ifn a
		     -1.0
		     (ifn b (* a -1) (- a (* a b)))) )
	  (to (ifn a
		   1.0
		   (ifn b a (+ a (* a b)))) )
	  (size (abs (- to from)) )
	  (tmp to) )

	(when (> from to)		; SWAP ITVL IF NEEDED
	    (setq to from)
	    (setq from tmp) )

	(+ from
	   (* size
	      (/ (random 10000.0) 10000.0))) ))

;------------------------------------------------------------------

(om::defmethod! find-chroma-file ((var t) (filename string))
  :icon 186
  (when (probe-file (getenv var))
    (make-pathname :directory (pathname-directory (getenv var)) :name filename)))
