(in-package :om)

;;;============================
;;; CR-MODEL
;;; data = a list of instances of chroma data classes (standard slot names, e.g. FREQ, AMP, etc.)
;;; timestruct = a temporal segmentation (list of dates)
;;;============================

(defclass! cr-model (control-data)
  ((modeltype :accessor modeltype :initarg :modeltype :initform 'FQL)
   (datatype :accessor datatype :initarg :datatype :initform '1TRC)
   (datasrc :accessor datasrc :initarg :datasrc :initform nil)
   ;;;(data :accessor data :initarg :data :initform nil)
   (time-struct :accessor time-struct :initarg :time-struct :initform nil))
  (:icon 654))


(defmethod get-slot-in-out-names ((self cr-model))
   (values '("self" "modeltype" "datatype" "datasrc" "time-struct")
           '(nil 'FQL '1TRC nil nil)
           '("object" "class name of the vertical structure" "SDIF analysis type" "source file" "temporal data")
           '(nil 
             ((1 (("Frequencies List" 'FQL)
                  ("Partials" 'PTL)
                  ("OM Chord" 'CHORD)
                  ))) 
             ((2 (("1TRC (additive analysis)" '1TRC) 
                  ("1MRK (chord-seq analysis)" '1MRK)
                  ("1HRM (harmonic analysis)" '1HRM)
                  ("1FOB (FOF analysis)" '1FOB)
                  ("1FQ0 (fundamental freq. analysis)" '1FQ0)) ))
             nil nil)
           ))

(defmethod cr-model-p ((self cr-model)) t)
(defmethod cr-model-p ((self t)) nil)

(defmethod convert-datasrc ((datasrc sdiffile))
  (filepathname datasrc))

(defmethod convert-datasrc ((datasrc t))
  datasrc)

;;; returns a list of markers from input 
(defmethod extract-time-struct ((self sdiffile))
  (remove-if (lambda (x) (< x 0)) (get-times self 0 "1MRK" nil nil nil)))

(defmethod extract-time-struct ((self list)) self)

(defmethod initialize-instance ((self cr-model) &rest initargs)
   (declare (ignore initargs)) 
   (call-next-method)
   (when (time-struct self)
     (setf (time-struct self) (extract-time-struct (slot-value self 'time-struct)))
     )
   )

(defmethod make-one-instance ((self cr-model) &rest slots-vals)
  (let ((rep (call-next-method)))
    (set-data rep)
    rep))

;;;=======================
;;; GET MODEL DATA
;;; see Model Processing for tranformations
(defmethod! model-data ((self list) &optional modif-func args)
  :icon 654
  (let ((rep (loop for elt in self collect (if elt (clone elt) nil))))
    (when modif-func
      (loop for i = 0 then (+ i 1)
         while (< i (length rep))
         do (let ((data (nth i rep)))
              (when data 
                (if args
                  (when (nth i args)
                    (setf (nth i rep) (apply modif-func (list data (nth i args)))))
                  (setf (nth i rep) (apply modif-func (list data))))
                ))))
    rep))

(defmethod! model-data ((self cr-model) &optional modif-func args)
  :icon 654
  (model-data (data self) modif-func args))



;;;=========================
;;; SPECIALIZED INSPECT
;; ((t1 ((ind freq amp phi) ... (ind freq amp phi))) 
;;  (t2 ((ind freq amp phi) ... (ind freq amp phi)))
;;  ...)
(defmethod get-analysis-data ((datatype (eql '1TRC)) datasrc beg end)
  (multiple-value-setq (data times) (getsdifdata datasrc 0 "1TRC" "1TRC" nil nil nil beg end))
  (mat-trans (list times data)))

;; ((t1 ((ind freq amp phi) ... (ind freq amp phi))) 
;;  (t2 ((ind freq amp phi) ... (ind freq amp phi)))
;;  ...)
(defmethod get-analysis-data ((datatype (eql '1HRM)) datasrc beg end)
  (multiple-value-setq (data times) (getsdifdata datasrc 0 "1HRM" "1HRM" nil nil nil beg end))
  (remove nil (mat-trans (list times data)) :key 'cdr))

;;
(defmethod get-analysis-data ((datatype (eql '1MRK)) datasrc beg end)
  (multiple-value-setq (data times) (getsdifdata datasrc 0 "1MRK" "1TRC" nil nil nil beg end))
  (mat-trans (list times data)))

;; 
(defmethod get-analysis-data ((datatype (eql '1FOB)) datasrc beg end)
  (multiple-value-setq (data times) (getsdifdata datasrc 0 "1FOB" "1FOF" nil nil nil beg end))
  (mat-trans (list times data)))

;; ((t1 f1) (t2 f2) ... (tn fn)) 
(defmethod get-analysis-data ((datatype (eql '1FQ0)) datasrc beg end)
  (multiple-value-setq (data times) (getsdifdata datasrc 0 "1FQ0" "1FQ0" 0 nil nil beg end))
  (mat-trans (list times (flat data))))

(defmethod get-data-times ((self pathname) datatype)
  (get-data-frames (load-sdif-file self) datatype))

(defmethod get-data-times ((self sdiffile) (datatype (eql '1TRC)))
  (get-times self 0 "1TRC" "1TRC" nil nil))

(defmethod get-data-times ((self sdiffile) (datatype (eql '1HRM)))
  (get-times self 0 "1HRM" "1HRM" nil nil))

(defmethod get-data-times ((self sdiffile) (datatype (eql '1MRK)))
  (get-times self 0 "1MRK" "1TRC" nil nil))

(defmethod get-data-times ((self sdiffile) (datatype (eql '1FOB)))
  (get-times self 0 "1FOB" "1FOF" nil nil))

(defmethod get-data-times ((self sdiffile) (datatype (eql '1FQ0)))
  (get-times self 0 "1FQ0" "1FQ0" nil nil))

;;;====================

(defmethod set-model-markers (markers datasrc (datatype (eql '1MRK)))
  (or markers (extract-time-struct datasrc)))

(defmethod set-model-markers (markers datasrc datatype)
  (let* ((f (get-data-times datasrc datatype))
         result)
    (when f
      (setf f (subseq f 0 (- (length f) 1)))
      (loop for mrk in markers
            do (let* ((diff_l (mapcar #'(lambda (x) (abs (- x mrk))) f))
                      (min (apply #'min diff_l))
                      (pos (position min diff_l)))
                 (push (nth pos f) result)))
      (setf result (remove-duplicates result :test #'eq))
      (sort result #'<)
      )))

;;; fills the data slot using file and datatype + time-struct
(defmethod set-data ((self cr-model))
  (cond ((consp (datasrc self))
         (when (>= (length (datasrc self)) (length (time-struct self)))
           (om-print "Warning some data was ignored because they were over the time structure"))
         (setf (data self) (first-n (datasrc self) (- (length (time-struct self)) 1)))
         (setf (modeltype self) (class-name (class-of (car (data self))))) 
         )
        ((datasrc self)
         (let ((ttt (cond ((or (time-struct self) (equal (datatype self) '1MRK)) (set-model-markers (time-struct self) (datasrc self) (datatype self))) 
                          ((setf tdata (get-data-times (datasrc self) (datatype self))) (list 0.0 (car (last tdata 2)))))))
           (when ttt (setf (time-struct self) ttt)))
         (setf (data self) (compute-model-data (datatype self) (modeltype self) (datasrc self) (time-struct self)))
         )
        (t (unless (time-struct self) (setf (time-struct self) '(0.0 1.0)))
           (setf (data self) (make-list (- (length (time-struct self)) 1))))))


;;;============================
;;; COMPUTE MODEL
;;;============================
;;; renvoie le nombre max d'lments
(defun max-chunk (data)
  (apply #'max (mapcar #'length data)))

(defun find-duration (n datalist)
  (let ((rep nil))
    (loop for frame in datalist
          do (when (member n (cadr frame) :key 'car :test '=)
               (pushr (car frame) rep)))
    (if rep
      (- (list-max rep) (list-min rep))
      0.0)))
  
(defun get-add-data (data numero colonne)
  (loop for frame in data
        collect (nth colonne (find numero frame :key #'car))))

;;;===============
;;; GENERAL METHOD
(defmethod compute-model-data (datatype modeltype datasrc markers) 
  (let ((data-list nil))
    (loop for (a b) on markers by #'cdr
          when b
          do (om-print (format nil "Segment : ~D -> ~D" a b))
          (let* ((analysisdata (get-analysis-data datatype datasrc a b))
                 modeldata)
            (when analysisdata
              (setf modeldata (get-model-data datatype modeltype analysisdata)))
            (push modeldata data-list)))
    (reverse data-list)))
       

;;;====================
;;; SPECIALIZED METHODS

(defmethod get-model-data ((datatype (eql '1TRC)) modeltype datalist &key (sort t) (weighed-avg t) (durmin 0.1))
  ;;; datalist = ((T1 (ind freq amp phi) ...) ... (Tn (ind freq amp phi) ...))
  (let (npartials freq_moyenne max_amp freq_fun amp_fun time_list triplets dur
                  (amps nil)
                  (freqs nil))
    (setf min-ind-partial (list-min (mapcar #'car (flat (mapcar #'cadr datalist) 1))))
    (setf max-ind-partial (list-max (mapcar #'car (flat (mapcar #'cadr datalist) 1))))
    (loop for i from min-ind-partial to max-ind-partial
            initially (setf amps nil freqs nil)
            do
            (setf freq_fun (get-add-data (mapcar #'cadr datalist) i 1))
            (setf amp_fun (get-add-data (mapcar #'cadr datalist) i 2))
            (setf dur (find-duration i datalist))
            (when (> dur durmin)
              (progn
                (setf triplets (mapcar #'list amp_fun freq_fun))
                (setf triplets (remove nil triplets :key 'car) ) ; ou delete ??
                (when (> (length triplets) 1)
                  ;amplitude moyenne
                  (setf amp_fun (mapcar #'first triplets))
                  (setf max_amp (apply #'max amp_fun))
                  (setf amp_fun (mapcar #'(lambda (x) (/ x max_amp)) amp_fun))
                  ;frequence moyenne
                  (setf freq_fun  (mapcar #'second triplets))
                  (setf freq_moyenne (if weighed-avg 
                                       (cr::moyenne_ponderee freq_fun amp_fun)
                                       (cr::moyenne freq_fun)))
                  (push freq_moyenne freqs)
                  (push max_amp amps)
                  ))))
      (if (not (null freqs))
        (progn (if sort (multiple-value-setq (freqs amps) (cr::sort-partiels freqs amps)))
               (make-model-obj modeltype :freq (nreverse freqs) :amp (nreverse amps))
               ))))



(defmethod get-model-data ((datatype (eql '1TRC)) (modeltype (eql 'PTL)) datalist &key (sort t) (weighed-avg t) (durmin 0.1))
  (let (npartials freq_moyenne max_amp freq_fun edel dur amp_fun time_list triplets subdata
                  (amps nil)
                  (edels nil)
                  (durs nil)
                  (freqs nil)
                  (amp_funs nil)
                  (freq_funs nil)
                  (init-time (car (car datalist))))
    (setf min-ind-partial (list-min (mapcar #'car (flat (mapcar #'cadr datalist) 1))))
    (setf max-ind-partial (list-max (mapcar #'car (flat (mapcar #'cadr datalist) 1))))
    (loop for i from min-ind-partial to max-ind-partial
          initially (setf amps nil freqs nil amp_funs nil freq_funs nil edels nil durs nil)
          do
          (setf freq_fun (get-add-data (mapcar #'cadr datalist) i 1))
          (setf amp_fun (get-add-data (mapcar #'cadr datalist) i 2))
          ;temps (X des fun)
          (setf time_list (mapcar #'car datalist))
          (setf triplets  (mapcar #'list amp_fun time_list freq_fun))
          ;; delete breakpoints absents de l'analyse
          (setf triplets (delete nil triplets :key 'car))
          (when (> (length triplets) 1) ;nombre minimum de points dans les bpf (pourrait tre 2)
            (setf time_list (mapcar #'second triplets))
            (setf dur  (- (last-elem time_list) (first time_list)))
            (when (> dur durmin) ;nombre minimum de points dans les bpf
              (progn
                ;amplitude maximum
                (setf amp_fun (mapcar #'first triplets))
                (setf max_amp (apply #'max amp_fun))
                ;normalisation des fonctions d'amplitudes ??
                ;(setf amp_fun (mapcar #'(lambda (x) (/ x max_amp)) amp_fun))
                ;frequence moyenne
                (setf freq_fun  (mapcar #'third triplets))
                (setf freq_moyenne (if weighed-avg 
                                     (cr::moyenne_ponderee freq_fun amp_fun)
                                     (cr::moyenne freq_fun))) 
                (push freq_moyenne freqs)
                (push max_amp amps)
                (setf freq_fun (mapcar #'(lambda (x) (cr::ratio->semitones (/ x freq_moyenne))) freq_fun))
                ;entry-delays
                (setf edel (- (first time_list) init-time))
                (push edel edels)
                (push dur durs)
                ;normalisation des fonctions en temps
                (setf freq_fun (cr::make_fun (om::flat (mapcar #'(lambda (x y) (list x y)) freq_fun time_list))))
                (cr::X-resc_fun freq_fun 0 1)
                (setf amp_fun (cr::make_fun (om::flat (mapcar #'(lambda (x y) (list x y)) amp_fun time_list))))
                (cr::X-resc_fun amp_fun 0 1)
                (push freq_fun freq_funs)
                (push amp_fun amp_funs)
                ))))
    (if (not (null freqs))
      (progn (if sort (multiple-value-setq (freqs amps edels durs freq_funs amp_funs) 
                        (cr::sort-partiels-and-everything freqs amps edels durs freq_funs amp_funs)))
             (make-instance 'cr::ptl
               :the-list (nreverse freqs) 
               :amplitudes (nreverse amps)                     
               :duration (- (car (last-elem datalist)) init-time)  
               :origtime init-time
               :entry-delays (nreverse edels)
               :durs (nreverse durs)
               :transp_funs (nreverse freq_funs)
               :amp_funs (nreverse amp_funs))
             ))))



(defmethod get-model-data ((datatype (eql '1HRM)) modeltype datalist &key (sort t) (weighed-avg t) (durmin 0.1))
  (get-model-data '1TRC modeltype datalist :sort sort :weighed-avg weighed-avg :durmin durmin))

(defmethod get-model-data ((datatype (eql '1MRK)) modeltype datalist &key (sort t) (weighed-avg t) (durmin 0.1))
  (let ((data (cadr (car datalist)))
        (amps nil) (freqs nil))
    (loop for p in data 
          initially (setf amps nil freqs nil)
          do
          (push (cadr p) freqs)
          (push (caddr p) amps)
          )
    (if (not (null freqs))
      (progn (if sort (multiple-value-setq (freqs amps) (cr::sort-partiels freqs amps)))
             (make-model-obj modeltype :freq (nreverse freqs) :amp (nreverse amps))
             ))))

(defmethod get-model-data ((datatype (eql '1FQ0)) modeltype datalist &key (sort t) (weighed-avg t) (durmin 0.1))
  ;;; datalist = ((t1 f1) ... (tn fn))
  (make-model-obj modeltype :freq (list (cr::moyenne (mapcar #'cadr datalist))))
  )



;;;====================
;;; SPECIALIZED TARGETS

(defmethod make-model-obj ((modeltype t) &key freq amp &allow-other-keys)
  nil)

(defmethod make-model-obj ((modeltype (eql 'FQL)) &key freq amp &allow-other-keys)
  (make-instance 'cr::fql 
    :the-list freq 
    :amplitudes amp))

(defmethod make-model-obj ((modeltype (eql 'CHORD)) &key freq amp &allow-other-keys)
  (let ((rep (make-instance 'chord 
               :lmidic (f->mc freq))))
    (when amp (setf (lvel rep) (om-round (om-scale amp 0 127))))
    rep))
  

(defmethod get-vps-freqs ((self t)) nil)
(defmethod get-vps-amps ((self t)) nil)

(defmethod get-vps-freqs ((self cr::fql))
  (cr::fql self))

(defmethod get-vps-amps ((self cr::fql))
  (cr::amplitudes self))

(defmethod get-vps-freqs ((self chord))
  (mc->f (lmidic self)))

(defmethod get-vps-amps ((self chord))
  (om-scale (lvel self) 0.0 1.0))

(defmethod get-vps-maxamp ((x t))
  (list-max (get-vps-amps x)))



;;;============================
;;; MODEL EDITOR
;;;============================

(defmethod Class-has-editor-p  ((self cr-model)) t )

(defmethod get-editor-class ((self cr-model)) 'modeleditor)

(omg-defclass modeleditor (EditorView object-editor) 
   ((maxfreq :initarg :maxfreq :accessor maxfreq :initform 4000.0)))

(defmethod get-panel-class ((Self modeleditor)) 'modelpanel)

(omg-defclass modelpanel (om-view) ())

(defmethod initialize-instance :after ((self modeleditor) &rest L) 
  (declare (ignore l))
  (let* ((Panel (om-make-view (get-panel-class self) 
                              :position (om-make-point 0 *titlebars-h*) 
                              :size (om-make-point (w self) (- (h self) (+ *titlebars-h* (get-control-h self))))
                              ;:bg-color *om-gray-color*
                              )))
    (om-add-subviews self panel)
    (setf (panel self) panel)))

(defmethod get-titlebar-class ((self modeleditor)) 'model-titlebar)
(omg-defclass model-titlebar (editor-titlebar) ())

(defmethod init-titlebar ((self model-titlebar))
  (let ((name (string+ (string (modeltype (object (om-view-container self)))) " Model")))
    (om-add-subviews self
                     (om-make-dialog-item 'om-static-text (om-make-point 10 4) 
                                          (om-make-point 200 ;;(om-string-size name *controls-font*) 
                                                         14)
                                        name
                                        :bg-color *editor-bar-color*
                                        :font *om-default-font2b*
                                        ))))

(defmethod update-subviews ((self modeleditor))
  (when (title-bar self)
     (om-set-view-size  (title-bar self) (om-make-point (w self) *titlebars-h*)))
  (om-set-view-size (panel self) (om-make-point (w self) (- (h self) *titlebars-h*)))
  ;(om-set-view-size  (control self) (om-make-point (w self) (get-control-h self)))
  ;(om-set-view-position (control self) (om-make-point 0 (- (h self) (get-control-h self))))
  (om-invalidate-view self))

(defmethod editor-null-event-handler :after ((self modeleditor))
  (unless (om-points-equal-p *mouse-window-event* (om-mouse-position  self))
     (setf *mouse-window-event* (om-mouse-position self))
     (when  (and (time-struct (object self))
                 (om-view-contains-point-p (panel self) *mouse-window-event*))
       (let ((p (om-convert-coordinates *mouse-window-event* self (panel self))))
         (om-with-focused-view (title-bar self)
           (om-with-fg-color (title-bar self) *editor-bar-color*
             (om-fill-rect (- (w self) 80) 2 80 22))
           (om-draw-string (- (w self) 80) 10 (format () "t = ~5f s" (/  (* (om-point-h p) (last-elem (time-struct (object self)))) (w self))))
           (om-draw-string (- (w self) 80) 21 (format () "f = ~5f Hz" (/  (* (- (h (panel self)) (om-point-v p)) (maxfreq self)) (h (panel self)))))
       )))))

(defvar *model-offset-click* nil)
(defvar *model-first-click* nil)
(defvar *model-last-click* nil)

(defmethod om-view-click-handler ((self modelpanel) where)
  (let* ((*model-offset-click* 0)
         (*model-first-click* where)
         (*model-last-click* where))
    (om-init-motion-functions self 'make-motion-model nil)))


(defmethod make-motion-model ((Self modelpanel) Where)
  (let*  ((first-mouse *model-first-click* )
          (old-mouse *model-last-click* )
          (new-mouse where) newrange)
   (setf *model-offset-click* (+ *model-offset-click* (- (om-point-v new-mouse) (om-point-v first-mouse) )))
   (setf newrange (+ (maxfreq (om-view-container self)) (* 10 delta)))
   (unless (or (> newrange 100000) (< newrange 50))
     (setf (maxfreq (om-view-container self)) newrange)
     (om-redraw-view self)
     (setf *model-last-click* new-mouse))))


(defmethod om-draw-contents ((self modelpanel))
  (let* ((model (object (om-view-container self))))
    (when (time-struct model)
      (let ((fact (/ (w self) (last-elem (time-struct model)))))
        (om-with-focused-view self
          (loop for (a b) on (time-struct model) by #'cdr
                for i = 0 then (+ i 1) do
                (let ((vps (nth i (data model))))
                  (om-draw-string (+ (round (* fact a)) 2) 10 (format nil "~4f" a))
                  (om-with-fg-color self (om-make-color 1 0.2 0.2)
                    (om-draw-line (round (* fact a)) 0 (round (* fact a)) (h self))
                    )
                (when (and b vps)
                  (loop for f in (get-vps-freqs vps) 
                        for i = 0 then (+ i 1) do 
                        (draw-nth-vps vps self i fact a b (h self) (maxfreq (om-view-container self)) (max-amp model))
                        ))
                )))))))


(defmethod draw-nth-vps ((self t) panel nth-elt time-fact begin-time end-time panel-h maxfreq  &optional (maxamp 1.0))
  t)

(defmethod draw-nth-vps ((self chord) panel nth-elt time-fact begin-time end-time panel-h maxfreq  &optional (maxamp 1.0))
  (let* ((f (nth nth-elt (get-vps-freqs self)))
         (amp (when (and (get-vps-amps self) (nth nth-elt (get-vps-amps self)))
                  (/ (nth nth-elt (get-vps-amps self)) maxamp)))
         (c (if amp (- 0.5 (* 0.5 amp)) 0.0))
         (freqh (round (- panel-h (* (/ f maxfreq) panel-h)))))
    (om-with-fg-color panel (om-make-color c c c)
      (om-draw-line (round (* time-fact begin-time)) freqh (round (* time-fact end-time)) freqh)))
  )

(defmethod draw-nth-vps ((self cr::FQL) panel nth-elt time-fact begin-time end-time panel-h maxfreq &optional (maxamp 1.0))
  (let* ((f (nth nth-elt (get-vps-freqs self)))
         (amp (when (and (get-vps-amps self) (nth nth-elt (get-vps-amps self)))
                  (/ (nth nth-elt (get-vps-amps self)) maxamp)))
         (c (if amp (- 0.5 (* 0.5 amp)) 0.0))
         (freqh (round (- panel-h (* (/ f maxfreq) panel-h)))))
    (om-with-fg-color panel (om-make-color c c c)
      (om-draw-line (round (* time-fact begin-time)) freqh (round (* time-fact end-time)) freqh)))
  )

(defmethod draw-nth-vps ((self cr::PTL) panel nth-elt time-fact begin-time end-time panel-h maxfreq &optional (maxamp 1.0))
  (let* ((f (nth nth-elt (get-vps-freqs self)))
         (amp (when (and (get-vps-amps self) (nth nth-elt (get-vps-amps self)))
                  (/ (nth nth-elt (get-vps-amps self)) maxamp)))
         (c (if amp (- 0.5 (* 0.5 amp)) 0.0))
         (freqh (round (- panel-h (* (/ f maxfreq) panel-h)))))
    ;(loop for ff in (nth nth-elt (cr::transp_funs self))
    ;      for aa in (nth nth-elt (cr::amp_funs self))
    ;      do
    ;      )
    (om-with-fg-color panel (om-make-color c c c)
      (om-draw-line (round (* time-fact (+ begin-time (nth nth-elt (cr::entry-delays self))))) freqh 
                    (round (* time-fact (+ begin-time (nth nth-elt (cr::entry-delays self)) (nth nth-elt (cr::durs self))))) freqh)))
    )


;;;=============================
;;; PERSISTANCE & SG DES MODELES

;;; ne pas recalculer data
(defmethod objfromobjs ((self cr-model) (type cr-model))
  (let ((new (make-instance 'cr-model 
               :datatype (datatype self)
               :modeltype (modeltype self)
               :time-struct (time-struct self))))
    (setf (data new) (data self))
    new))

(defmethod objfromobjs ((self cr-model) (type sdiffile))
  (let ((path (om-choose-new-file-dialog :prompt "Choose a file name for saving your model")))
    (save-model self path)))

 (defmethod objfromobjs ((self sdiffile) (type cr-model))
   (let ((new (make-instance 'cr-model))
         (tlist nil)
         (datalist nil)
         (ptrfile (dynamic-open self)))
       (sdif::SdifFReadGeneralHeader ptrfile)
       (sdif::SdifFReadAllASCIIChunks ptrfile)
       (loop for item in (framesdesc self) do
             (when (equal "1MRK" (car item))
               (push (nth 1 item) tlist)
               (sdif-set-pos ptrfile (nth 3 item))
               (setf mlist (nth 4 item))
               (loop for mat in mlist do
                     (cond ((equal "1BEG" (car mat)) (setf bmat mat))
                           ((equal "1END" (car mat)) (setf emat mat))
                           ((equal "1TRC" (car mat)) (setf pmat mat))
                           (t nil)))
               (when pmat 
                 (let ((flist nil) (alist nil))
                   ;;; a parameter matrix :
                   ;;; cherche les notes dans tmplist et set pitch et velocity
                   (sdif-read-headers ptrfile (nth 3 item) (fifth pmat))
                   (loop for i = 0 then (+ i 1) while (< i (second pmat)) do
                         (sdif::SdifFReadOneRow ptrfile)
                         (let* ((freq (sdif::SdifFCurrOneRowCol ptrfile 2))
                                (amp (sdif::SdifFCurrOneRowCol ptrfile 3)))
                           (pushr freq flist)
                           (pushr amp alist)
                           ))
                   (when flist (push (make-instance 'cr::fql :the-list flist :amplitudes alist) datalist))
                   ))
               (setf bmat nil)
               (setf emat nil)
               (setf pmat nil)
               )
             )
       (dynamic-close self ptrfile)
       (setf (time-struct new) (reverse tlist))
       (setf (data new) (reverse datalist))
       new))


(defmethod save-model ((self cr-model) filename)
  (let ((outfile (sdif-open-file (om-path2cmdpath (pathname filename)) 1)))
    (sdif-write-header outfile (list (make-instance 'sdiftype 
                                       :struct 'F :signature "1MRK" 
                                       :description '(("1TRC" "model_data")))))
    (loop for time in (time-struct self) 
          for i = 0 then (+ i 1) do
          (let ((frame (make-instance 'sdifframe 
                         :signature "1MRK"
                         :ftime time
                         :streamId 0))
                (data (nth i (data self)))
                (matrix (make-instance 'soft-sdifmatrix :signature "1TRC" 
                                       :num-elts 0 :num-fields 4)))
            (when data
              (setf (data matrix) 
                    (flat (loop for f in (cr::fql data) 
                          for n = 0 then (+ n 1) collect 
                          (list n f (or (nth n (cr::amplitudes data)) 1.0) 0))))
              (setf (num-elts matrix) (length (cr::fql data))))
            (setf (lmatrix frame) (list matrix))
            (save-sdif frame outfile)
            ))
    filename))



