(in-package :osc)
(export  '(open-osc-out-stream close-osc-stream write-osc-msg write-osc-bundle start-osc-server stop-osc-server decode-msg-or-bundle))

;; the representation of an osc msg is a list of string or fixnum objects
;; a bundle is a list of such. Upon reception, decode-bundle gives a list with a time-tag as its first element
;; does not handle floats and int are limited to 2^31

;; as udpsend and udpreceive tend to replace OpenSoundControl in max, there is a simplewrite-osc-msg
;; in addition to write-osc-bundle

(defun write-osc-msg (message osc-stream)
  (write-sequence (apply #'osc:encode-message message) osc-stream)
  (force-output osc-stream))

(defun write-osc-bundle (bundle osc-stream &optional (time-tag :now))
  (write-sequence (osc:encode-bundle bundle time-tag) osc-stream)
  (force-output osc-stream))

(defun open-osc-out-stream (host port)
  (comm:open-udp-stream host port :direction :output :element-type '(unsigned-byte 8)))

(defun close-osc-stream (osc-stream)
  (close osc-stream))

(defun start-osc-server (host port &optional (function #'print-incoming-osc-msg))
 (comm:start-udp-server :address host :service port :function function
                        :process-name (format nil "OSC receiver on ~S ~S" host port)) )

(defun stop-osc-server (server)
  (comm:stop-udp-server  server :wait t))

(defun print-incoming-osc-msg (msg host)
  (format cl-user::*stdout* "Osc received...... ~a from ~a~%" (decode-msg-or-bundle msg) host)
  nil)

(defun decode-msg-or-bundle (msg-or-bundle)
  (if (= (elt msg-or-bundle 0) 35)
      (osc:decode-bundle msg-or-bundle)
    (osc:decode-message msg-or-bundle)))


;; modifications to osc external library -----------------------------------------------
;; just add force-string for symbols

(defun force-string (data) (string-downcase (string data)))

(defun encode-message (address &rest data)
  "encodes an osc message with the given address and data."
  (concatenate '(vector (unsigned-byte 8))
	       (encode-address (force-string  address))
	       (encode-typetags data)
	       (encode-data data)))

(defun encode-address (address)
  (cat (map 'vector #'char-code address) 
       (string-padding address)))

(defun encode-typetags (data)
  "creates a typetag string suitable for the given data.
  valid typetags according to the osc spec are ,i ,f ,s and ,b
  non-std extensions include ,{h|t|d|S|c|r|m|T|F|N|I|[|]}
                             see the spec for more details. ..

  NOTE: currently handles the following tags 
   i => #(105) => int32
   f => #(102) => float
   s => #(115) => string 
   b => #(98)  => blob
  and considers non int/float/string data to be a blob." 

  (let ((lump (make-array 0 :adjustable t 
			  :fill-pointer t)))
    (macrolet ((write-to-vector (char)
                 `(vector-push-extend
                   (char-code ,char) lump)))
      (write-to-vector #\,)
      (dolist (x data) 
        (typecase x
          (integer (write-to-vector #\i))
          (float (write-to-vector #\f))
          (simple-string (write-to-vector #\s))
          (symbol (write-to-vector #\s))
	  (t (write-to-vector #\b)))))
    (cat lump
         (pad (padding-length (length lump))))))     
		  
(defun encode-data (data)
  "encodes data in a format suitable for an OSC message"
  (let ((lump (make-array 0 :adjustable t :fill-pointer t)))
    (macrolet ((enc (f)
                 `(setf lump (cat lump (,f x)))))
      (dolist (x data) 
        (typecase x
          (integer (enc encode-int32)) 
          (float (enc encode-float32)) 
          (simple-string (enc encode-string))
          (symbol (setf lump (cat lump (encode-string (force-string x)))))
	  (t (enc encode-blob))))
      lump)))

;---------------------------------------------------------------------------------


#|


;; start a server for processing incoming osc messages
;; by default, the server will just print the incoming messages on cl-user::*stdout*
;; you can bind cl-user::*stdout* to *standard-output*

(setf os (start-osc-server "localhost" 3500 ))

;; create a stream for writing out osc msgs
(setf outs (open-osc-out-stream "localhost" 2000))

;; write message
(write-osc-msg '("/toto" 100 200 300 "titi" 2000) outs)
(write-osc-bundle '(("/toto" 100 200 300 "titi" 2000) ("/toto" 100 200 300 "titi" 2000) ("/toto" 100 200 300 "titi" 2000)) outs)

;; close osc out stream
(close-osc-stream outs)

;; stop osc incoming server
(stop-osc-server os)


;; this is how you would define your own osc message processing on input

(defun handle-osc-msg (msg host)
  (setf *curmsg* (decode-msg-or-bundle msg))
  ;;do some  processing on *curmsg*
  (format cl-user::*stdout* "~a~%" (mapcar #'type-of *curmsg*))
  ;; return nil so the udp server does not send back an message to the sender
  nil)

(setf os (start-osc-server "localhost" 3500 #'handle-osc-msg))

(stop-osc-server os)


|#



